import pm4py
from pm4py.algo.simulation.playout.petri_net import algorithm as simulator
from tqdm import tqdm


def simulate_model(petri_net, im, fm, size, log=None):
    """
    creates a log of size <no_traces> adhering to the behaviour described by the provided net. If a log is provided
    transition probabilities are chosen according to the stochastic map built from the provided log.
    :param size: size of the returned event log
    :param petri_net: a petri net
    :param im: initial marking of the petri net
    :param fm: final marking of the petri net
    :param log: event log determining transition probabilities
    :return: event log
    """
    log = pm4py.algo.simulation.playout.petri_net.algorithm.apply(petri_net, im, fm, {'noTraces': size,  'maxTraceLength':1000},
                                                                  variant=pm4py.algo.simulation.playout.petri_net.algorithm.Variants.BASIC_PLAYOUT)
    for i in range(len(log)):
        while len(log[i]) == 0:
            tr = pm4py.algo.simulation.playout.petri_net.algorithm.apply(petri_net, im, fm, {'noTraces': 1, 'maxTraceLength':1000},
                                                                                      variant=pm4py.algo.simulation.playout.petri_net.algorithm.Variants.BASIC_PLAYOUT)[0]
            log[i] = tr
        if len(log[i]) == 0:
            print("ZEROLENGTH TRACE")
    return log



import math
import statistics

import numpy as np

from special4pm.estimation.metrics import get_singletons, get_doubletons, get_total_species_count, \
    hill_number_asymptotic, completeness, \
    coverage


def generate_bootstrap_samples_abundance(reference_sample, n):
    s_obs = len(reference_sample)
    #print(reference_sample.values)
    sample_size = get_total_species_count(reference_sample)

    f_1 = get_singletons(reference_sample)
    f_2 = get_doubletons(reference_sample)
    f_0 = 0
    if f_2>0:
        f_0 = ((sample_size-1)/sample_size) * f_1**2 / (2*f_2)
    else:
        f_0 = ((sample_size-1)/sample_size) * f_1 * (f_1-1) / 2
    f_0 = math.ceil(f_0)
    probabilities = []
    c = get_c_n_abundance(reference_sample, sample_size)
    factor = factor_abundance(reference_sample, sample_size, c)


    for x_i in reference_sample.values():
        adapted_p = (x_i / sample_size) * (1 - factor * ((1 - (x_i / sample_size)) ** sample_size))
        probabilities.append(adapted_p)
    [probabilities.append((1 - c) / f_0) for i in range(0, f_0)]

    rng = np.random.default_rng()
    bs_samples = []
    #for i in range(n):
    #    selected_species = choice([x for x in range(0, s_obs + f_0)], sample_size, p=probabilities)

    #    unique, counts = np.unique(selected_species, return_counts=True)
    for x in rng.multinomial(sample_size, probabilities, size=n):
        bs_samples.append( dict(zip(range(len(x)), x)))

    return bs_samples


def get_bootstrap_ci_incidence(reference_sample, sample_size, no_samples):
    samples = generate_bootstrap_samples_incidence(reference_sample, sample_size, no_samples)
    d0 = [hill_number_asymptotic(0, sample, sample_size, abundance=False) for sample in samples]
    d1 = [hill_number_asymptotic(1, sample, sample_size, abundance=False) for sample in samples]
    d2 = [hill_number_asymptotic(2, sample, sample_size, abundance=False) for sample in samples]
    c0 = [completeness(sample) for sample in samples]
    c1 = [coverage(sample, sample_size) for sample in samples]

    return [statistics.stdev(x)*1.96 for x in (d0,d1,d2,c0,c1)]


def generate_bootstrap_samples_incidence(reference_sample, sample_size, no_bs_samples):
    #get bootstrap distribution of species
    f_0 = 0

    f_1 = get_singletons(reference_sample)
    f_2 = get_doubletons(reference_sample)

    if f_2>0:
        f_0 = ((sample_size-1)/sample_size) * f_1**2 / (2*f_2)
    else:
        f_0 = ((sample_size-1)/sample_size) * f_1 * (f_1-1) / 2
    f_0=math.ceil(f_0)

    probabilities = []
    c = get_c_n_incidence(reference_sample, sample_size)
    u = sum(reference_sample.values())

    factor = factor_incidence(reference_sample, sample_size, u, c)

    for y_i in reference_sample.values():
        adapted_p = (y_i / sample_size) * (1 - factor * (1 - (y_i / sample_size) ** sample_size))
        probabilities.append(adapted_p)
    [probabilities.append((u/sample_size) * (1 - c) / f_0) for i in range(0, f_0)]

    #generate all samples
    bootstrap_samples = [{} for _ in range(no_bs_samples)]
    #for n in range(0,sample_size):
    for s,p in enumerate(probabilities):
        species_counts = np.random.default_rng().binomial(n=sample_size, p=p, size=no_bs_samples)
        for x in range(no_bs_samples):
            if species_counts[x]==0:
                continue
            bootstrap_samples[x][s]=species_counts[x]

    return bootstrap_samples

def generate_bootstrap_estimates_incidence(reference_sample, sample_size):
    #get bootstrap distribution of species
    f_0 = 0

    f_1 = get_singletons(reference_sample)
    f_2 = get_doubletons(reference_sample)

    if f_2>0:
        f_0 = ((sample_size-1)/sample_size) * f_1**2 / (2*f_2)
    else:
        f_0 = ((sample_size-1)/sample_size) * f_1 * (f_1-1) / 2
    f_0=math.ceil(f_0)

    probabilities = []
    c = get_c_n_incidence(reference_sample, sample_size)
    u = sum(reference_sample.values())

    factor = factor_incidence(reference_sample, sample_size, u, c)

    for y_i in reference_sample.values():
        adapted_p = (y_i / sample_size) * (1 - factor * (1 - (y_i / sample_size) ** sample_size))
        probabilities.append(adapted_p)
    [probabilities.append((u/sample_size) * (1 - c) / f_0) for i in range(0, f_0)]

    #generate all samples
    bootstrap_sample = {}
    #for n in range(0,sample_size):
    for s,p in enumerate(probabilities):
        species_count = np.random.default_rng().binomial(n=sample_size, p=p)
        bootstrap_sample[s]=species_count

    #collect estimates

    #return tuple
    return bootstrap_sample


def factor_abundance(reference_sample, n, c):
    if c == 1:
        return 0
    return (1 - c) / (sum([(x_i / n) * (1 - (x_i / n))**n for x_i in reference_sample.values()]))


def factor_incidence(reference_sample, n, u, c):
    if c == 1:
        return 0
    return ((u / n) * (1 - c)) / (sum([(x_i / n) * (1 - (x_i / n) ** n) for x_i in reference_sample.values()]))


def get_c_n_abundance(reference_sample, n):
    f_1 = get_singletons(reference_sample)
    f_2 = get_doubletons(reference_sample)

    if f_2 > 0:
        return 1 - (f_1 / n) * (((n - 1) * f_1) / ((n - 1) * f_1 + 2 * f_2))
    else:
        return 1 - (f_1 / n) * (((n - 1) * (f_1 - 1)) / ((n - 1) * (f_1 - 1) + 2))


def get_c_n_incidence(reference_sample, n):
    u = sum(reference_sample.values())
    f_1 = get_singletons(reference_sample)
    f_2 = get_doubletons(reference_sample)
    if f_2 > 0:
        return 1-(f_1/ u) * (((n-1)*f_1)/((n-1)*f_1+2*f_2))
    else:
        return 1-(f_1/ u) * (((n-1)*(f_1-1))/((n-1)*(f_1-1)+2))


import math
from datetime import datetime, timedelta

import pm4py

def retrieve_species_n_gram(trace, n):
    """
    TODO: this return a list of exactly n-grams, i.e. if a sequence is less than n-grams, an empty list is returned
    :param trace:
    :param n:
    :return:
    """
    #if len(trace) < n:
    #    return ["NULL"]
    events = [x['concept:name'] for x in trace]
    #if len(events) <n:
    #    [events.append("NULL") for _ in range(n-len(events))]
    #if n >= 2:
    #    events.insert(0, "START")
    #    events.append("END")
    return [",".join(events[x:x + n]) for x in range(0, len(events) - n+1)]


def retrieve_species_trace_variant(trace):
    return [",".join([x["concept:name"] for x in trace])]


def retrieve_timed_activity(trace, interval_size):
    l=[]
    if "lifecycle:transition" in trace[0]:
        open_set = []
        for ide,e in enumerate(trace):
            if "start" in e["lifecycle:transition"].lower():
                open_set.append(e)
            elif "complete" in e["lifecycle:transition"].lower():
                found = False
                for idx, x in enumerate(open_set):
                    if x["concept:name"] == e["concept:name"]:
                        time = trace[ide]["time:timestamp"] - trace[idx]["time:timestamp"]
                        t = interval_size * math.ceil((time.total_seconds() / 60 / 60) / interval_size)
                        l.append(e["concept:name"] + "_" + str(t))
                        open_set.pop(idx)
                        found = True
                        break
                if not found:
                    if ide == 0:
                        l.append(e["concept:name"]+"_"+str(interval_size))
                    else:
                        time = trace[ide]["time:timestamp"] - trace[ide-1]["time:timestamp"]
                        t = interval_size * math.ceil((time.total_seconds() / 60 / 60) / interval_size)
                        l.append(e["concept:name"] + "_" + str(t))
    else:
        if len(trace) == 1:
            return [trace[0]["concept:name"] + "_0"]
        t = sorted(trace, key=lambda d: d['time:timestamp'])
        for idx,e in enumerate(trace):
            if idx == 0:
                l.append(e["concept:name"]+"_0")
            else:
                time = trace[idx]["time:timestamp"] - trace[idx-1]["time:timestamp"]
                t = interval_size * math.ceil((time.total_seconds()/60/60)/interval_size)
                l.append(e["concept:name"]+"_"+str(t))
    #print(l, len(trace))
    return l

def retrieve_timed_activity_exponential(trace):
    l = []
    if len(trace) == 1:
        return [trace[0]["concept:name"] + "_2"]
    if "lifecycle:transition" in trace[0]:
        open_set = []
        for ide,e in enumerate(trace):
            if "start" in e["lifecycle:transition"].lower():
                open_set.append(e)
            elif "complete" in e["lifecycle:transition"].lower():
                found = False
                for idx, x in enumerate(open_set):
                    if x["concept:name"] == e["concept:name"]:
                        time = trace[ide]["time:timestamp"] - trace[idx]["time:timestamp"]
                        t = time.total_seconds() / 60 / 60
                        if t < 1:
                            l.append(e["concept:name"] + "_2")
                        else:
                            l.append(e["concept:name"] + "_" + str(2 * math.ceil(math.log(t, 2))))
                        open_set.pop(idx)
                        found = True
                        break
                if not found:
                    if ide == 0:
                        l.append(e["concept:name"]+"_2")
                    else:
                        time = trace[ide]["time:timestamp"] - trace[ide-1]["time:timestamp"]
                        t = time.total_seconds() / 60 / 60
                        if t < 1:
                            l.append(e["concept:name"] + "_2")
                        else:
                            l.append(e["concept:name"] + "_" + str(2 * math.ceil(math.log(t, 2))))
    else:
        t = sorted(trace, key=lambda d: d['time:timestamp'])
        for idx, e in enumerate(t):
            if idx == 0:
                l.append(e["concept:name"] + "_0")
            else:
                time = trace[idx]["time:timestamp"] - trace[idx-1]["time:timestamp"]
                t = time.total_seconds() / 60/60
                if t < 1:
                    l.append(e["concept:name"] + "_2")
                else:
                    l.append(e["concept:name"] + "_" + str(2 * math.ceil(math.log(t,2))))
    return l


#log = pm4py.read_xes("logs/Sepsis_Cases_-_Event_Log.xes", return_legacy_log_object=True)
#for x in log:
#    retrieve_timed_activity_exponential(x)


from matplotlib import pyplot as plt

plt.rcParams['figure.figsize'] = [5, 4]
plt.rcParams['xtick.labelsize'] = 20
plt.rcParams['ytick.labelsize'] = 20

plt.plot([36,34.2,33.0], label="Equal", linestyle="--")
plt.plot([36,27.6,22.4], label="Added", linestyle="--")
plt.plot([25,18.4,14.7], label="Base", linestyle="--")
plt.plot([16,7.7,4.7], label="Removed", linestyle="--")

plt.xticks([0,1,2],[0,1,2])


#ax1.set_title(names[0], fontsize=24)
#ax1.set_xlabel("Species Rank", fontsize=20)
#ax1.set_ylabel("Occurences", fontsize=20)
#ax1.set_xticks([0, no_species - 1])
#ax1.set_xticklabels([1, no_species])
#plt.yticks([0, max(estimations[0].reference_sample_abundance.values()) - 1],
#           [0, max(estimations[0].reference_sample_abundance.values()) - 1])
plt.xlabel("Order q", fontsize=20)
plt.ylabel("Hill Number", fontsize=20)

plt.legend()
plt.tight_layout()
plt.savefig("figures/drift_eval_diversity_profiles.pdf", format="pdf")
#plt.show()
plt.close()
# plt.show()

plt.rcParams['figure.figsize'] = [5, 4]
plt.rcParams['xtick.labelsize'] = 20
plt.rcParams['ytick.labelsize'] = 20

plt.plot([717, 45.8, 22.5], label="BPI-2018", linestyle="--")
plt.plot([618, 30.4, 18.7], label="BPI-2019", linestyle="--")
plt.plot([164, 39.6, 23.9], label="Sepsis", linestyle="--")
plt.plot([150, 36.9, 18.8], label="BPI-2012", linestyle="--")

plt.xticks([0,1,2],[0,1,2])
plt.yscale("log")

#ax1.set_title(names[0], fontsize=24)
#ax1.set_xlabel("Species Rank", fontsize=20)
#ax1.set_ylabel("Occurences", fontsize=20)
#ax1.set_xticks([0, no_species - 1])
#ax1.set_xticklabels([1, no_species])
#plt.yticks([0, max(estimations[0].reference_sample_abundance.values()) - 1],
#           [0, max(estimations[0].reference_sample_abundance.values()) - 1])
plt.xlabel("Order q", fontsize=20)
plt.ylabel("Hill Number", fontsize=20)
plt.ylim(0,1000)
plt.legend()
plt.tight_layout()
plt.savefig("figures/real_eval_diversity_profiles.pdf", format="pdf")
#plt.show()
plt.close()
# plt.show()


import math
from src.estimation.metrics import get_singletons


def extrapolate_richness_abundance(reference_sample, sample_size, richness, data_points=100):
    s_obs = len(reference_sample)
    f_0 = richness - s_obs
    f_1 = get_singletons(reference_sample)
    m_size = math.floor(sample_size / data_points)
    values = []
    locations = []
    m = 0
    while m + m_size < sample_size:
        m = m + m_size
        #print("Richness Extrapolation - M+m=" + str(sample_size + m))
        if f_0 == 0:
            values.append(s_obs)
        else:
            values.append(s_obs + f_0 * (1 - (1 - (f_1 / (sample_size * f_0 + f_1))) ** m))
        locations.append(sample_size + m)

    m =  sample_size
    if f_0 == 0:
        values.append(s_obs)
    else:
        values.append(s_obs + f_0 * (1 - (1 - (f_1 / (sample_size * f_0 + f_1))) ** m))
    locations.append(sample_size + m)
    return values, locations


# structurally equivalent to abundance case
def extrapolate_richness_incidence(reference_sample, sample_size, richness, data_points=100):
    return extrapolate_richness_abundance(reference_sample, sample_size, richness, data_points)


def extrapolate_shannon_entropy_abundance(reference_sample, sample_size, asymp_entr, data_points=100):
    m_size = math.floor(sample_size / data_points)
    asymp_entr = math.log(asymp_entr)
    values = []
    locations = []
    m = 0
    while m + m_size < sample_size:
        m = m + m_size
        #print("Shannon Entropy Extrapolation - M+m=" + str(sample_size + m))

        first = sample_size / (sample_size + m)
        second = (sum([- (x_i / sample_size) * math.log(x_i / sample_size) for x_i in reference_sample.values()]))
        third = m / (sample_size + m) * asymp_entr
        values.append(math.exp(first * second + third))
        locations.append(sample_size + m)

    m = sample_size
    first = sample_size / (sample_size + m)
    second = (sum([- (x_i / sample_size) * math.log(x_i / sample_size) for x_i in reference_sample.values()]))
    third = m / (sample_size + m) * asymp_entr
    values.append(math.exp(first * second + third))
    locations.append(sample_size + m)
    return values, locations


# structurally equivalent to abundance case with adapted sample size
def extrapolate_shannon_entropy_incidence(reference_sample, sample_size, asymp_entr, data_points=100):
    u = sum(list(reference_sample.values()))

    m_size = math.floor(sample_size / data_points)
    asymp_entr = math.log(asymp_entr)
    values = []
    locations = []
    m = 0
    while sample_size + m + m_size < 2 * sample_size:
        m = m + m_size
        #print("Shannon Entropy Extrapolation - M+m=" + str(sample_size + m))

        first = sample_size / (sample_size + m)
        second = (sum([- (y_i / u) * math.log(y_i / u) for y_i in reference_sample.values()]))
        third = m / (sample_size + m) * asymp_entr
        values.append(math.exp(first * second + third))
        locations.append(sample_size + m)

    m = 2 * sample_size
    first = sample_size / (sample_size + m)
    second = (sum([- (y_i / u) * math.log(y_i / u) for y_i in reference_sample.values()]))
    third = m / (sample_size + m) * asymp_entr
    values.append(math.exp(first * second + third))
    locations.append(m)
    return values, locations


def extrapolate_simpson_diversity_abundance(reference_sample, sample_size, data_points=100):
    m_size = math.floor(sample_size / data_points)

    values = []
    locations = []
    m = 0
    while sample_size + m + m_size < 2 * sample_size:
        m = m + m_size
        #print("Simpson Diversity Extrapolation - M+m=" + str(sample_size + m))

        first = 1 / (sample_size + m)
        second = (sample_size + m - 1) / (sample_size + m)
        third = sum([(x_i * (x_i - 1)) / (sample_size * (sample_size - 1)) for x_i in reference_sample.values()])
        values.append(1 / (first + second * third))
        locations.append(sample_size + m)

    m = 2 * sample_size
    first = 1 / (sample_size + m)
    second = (sample_size + m - 1) / (sample_size + m)
    third = sum([(x_i * (x_i - 1)) / (sample_size * (sample_size - 1)) for x_i in reference_sample.values()])
    values.append(1 / (first + second * third))
    locations.append(m)
    return values, locations


def extrapolate_simpson_diversity_incidence(reference_sample, sample_size, data_points=100):
    m_size = math.floor(sample_size / data_points)
    u = sum(list(reference_sample.values()))

    values = []
    locations = []
    m = 0
    while sample_size + m + m_size < 2 * sample_size:
        m = m + m_size
        #print("Simpson Diversity Extrapolation - M+m=" + str(sample_size + m))

        first = 1 / (sample_size + m)
        second = 1 / (u / sample_size)
        third = (sample_size + m - 1) / (sample_size + m)
        fourth = sum([(y_i * (y_i - 1)) / (u ** 2 * (1 - (1 / sample_size))) for y_i in reference_sample.values()])
        values.append(1 / ((first * second) + third * fourth))
        locations.append(sample_size + m)

    m = 2 * sample_size

    first = 1 / (sample_size + m)
    second = 1 / (u / sample_size)
    third = (sample_size + m - 1) / (sample_size + m)
    fourth = sum([(y_i * (y_i - 1)) / (u ** 2 * (1 - (1 / sample_size))) for y_i in reference_sample.values()])
    values.append(1 / ((first * second) + third * fourth))
    locations.append(m)
    return values, locations



from typing import Callable

import pandas as pd
import pm4py
from deprecation import deprecated
from pandas import DataFrame
from pm4py.objects.log.obj import EventLog, Trace
from special4pm.bootstrap import bootstrap
from tqdm import tqdm

from special4pm.estimation.metrics import get_singletons, get_doubletons, completeness, coverage, \
    sampling_effort_abundance, sampling_effort_incidence, hill_number_asymptotic, entropy_exp, simpson_diversity


# TODO enum for proper key access
# TODO redo print to be r-like table of current values or history of values
# TODO differentiate between abundance and incidence
# TODO maybe remove MetricManager and have class extend hash map directly
# TODO move bootstrap out of here
# TODO dataFrame incorporate all information

class MetricManager(dict):
    # TODO convert to dataclass
    """
    Manages metrics for abundance and incidence models.
    """
    def __init__(self, d0: bool, d1: bool, d2: bool, c0: bool, c1: bool, l_n: list) -> None:
        # reference sample stats
        super().__init__()
        self.reference_sample_abundance = {}
        self.reference_sample_incidence = {}

        self.incidence_current_total_species_count = 0
        self.abundance_current_total_species_count = 0
        self.incidence_sample_size = 0
        self.abundance_sample_size = 0
        self.current_co_occurrence = 0
        self.empty_traces = 0

        self["abundance_no_observations"] = [0]
        self["incidence_no_observations"] = [0]
        self["abundance_sum_species_counts"] = [0]
        self["incidence_sum_species_counts"] = [0]
        self["degree_of_co_occurrence"] = [0]
        self["abundance_singletons"] = [0]
        self["incidence_singletons"] = [0]
        self["abundance_doubletons"] = [0]
        self["incidence_doubletons"] = [0]
        if d0:
            self["abundance_sample_d0"] = [0]
            self["incidence_sample_d0"] = [0]
            self["abundance_estimate_d0"] = [0]
            self["incidence_estimate_d0"] = [0]
            self["incidence_estimate_d0_ci"] = [-1]

        if d1:
            self["abundance_sample_d1"] = [0]
            self["incidence_sample_d1"] = [0]
            self["abundance_estimate_d1"] = [0]
            self["incidence_estimate_d1"] = [0]
            self["incidence_estimate_d1_ci"] = [-1]

        if d2:
            self["abundance_sample_d2"] = [0]
            self["incidence_sample_d2"] = [0]
            self["abundance_estimate_d2"] = [0]
            self["incidence_estimate_d2"] = [0]
            self["incidence_estimate_d2_ci"] = [-1]

        if c0:
            self["abundance_c0"] = [0]
            self["incidence_c0"] = [0]
            self["incidence_c0_ci"] = [-1]

        if c1:
            self["abundance_c1"] = [0]
            self["incidence_c1"] = [0]
            self["incidence_c1_ci"] = [-1]

        for l in l_n:
            self["abundance_l_" + str(l)] = [0]
            self["incidence_l_" + str(l)] = [0]
            self["incidence_l_" + str(l)+"_ci"] = [-1]


class SpeciesEstimator:
    """
    A class for the estimation of diversity and completeness profiles of trace-based species definitions
    """

    def __init__(self, d0: bool = True, d1: bool = True, d2: bool = True, c0: bool = True,
                 c1: bool = True,
                 l_n: list = [.9, .95, .99], no_bootstrap_samples: int = 0, step_size: int | None = None):
        """
        :param d0: flag indicating if D0(=species richness) should be included
        :param d1: flag indicating if D1(=exponential Shannon entropy) should be included
        :param d2: flag indicating if D2(=Simpson diversity index) should be included
        :param c0: flag indicating if C0(=completeness) should be included
        :param c1: flag indicating if C1(=coverage) should be included
        :param l_n: list of desired completeness values for estimation additional sampling effort
        :param step_size: the number of added traces after which the profiles are updated. Use None if
        """
        # TODO add differentiation between abundance and incidence based data
        self.include_abundance = True
        self.include_incidence = True

        self.include_d0 = d0
        self.include_d1 = d1
        self.include_d2 = d2

        self.include_c0 = c0
        self.include_c1 = c1

        self.no_bootstrap_samples = no_bootstrap_samples

        self.l_n = l_n

        self.step_size = step_size

        self.metrics = {}
        self.species_retrieval = {}

        self.current_obs_empty = False

    def register(self, species_id: str, function: Callable) -> None:
        self.species_retrieval[species_id] = function
        self.metrics[species_id] = MetricManager(self.include_d0, self.include_d1, self.include_d2, self.include_c0,
                                                 self.include_c1, self.l_n)

    def add_bootstrap_ci(self, sample_size):
        #print("Adding Bootstrapping Confidence Intervals")
        for species_id in self.metrics.keys():
            ci=(bootstrap.get_bootstrap_ci_incidence(self.metrics[species_id].reference_sample_incidence,
                                                       self.metrics[species_id].incidence_sample_size - self.metrics[species_id].empty_traces,
                                                       sample_size))
            self.metrics[species_id]["incidence_estimate_d0_ci"][-1] = ci[0]
            self.metrics[species_id]["incidence_estimate_d1_ci"][-1] = ci[1]
            self.metrics[species_id]["incidence_estimate_d2_ci"][-1] = ci[2]
            self.metrics[species_id]["incidence_c0_ci"][-1] = ci[3]
            self.metrics[species_id]["incidence_c1_ci"][-1] = ci[4]

        #    print(species_id + "Estimates + CI")
            d0 = self.metrics[species_id]["incidence_estimate_d0"][-1]
        #    print("D0: " + str(d0) + " (" + str(d0 - ci[0]), str(d0 + ci[0]) + ")")
            d1 = self.metrics[species_id]["incidence_estimate_d1"][-1]
        #    print("D1: " + str(d1) + " (" + str(d1 - ci[1]), str(d1 + ci[1]) + ")")
            d2 = self.metrics[species_id]["incidence_estimate_d2"][-1]
        #    print("D2: " + str(d2) + " (" + str(d2 - ci[2]), str(d2 + ci[2]) + ")")
            c0 = self.metrics[species_id]["incidence_c0"][-1]
        #    print("C0: " + str(c0) + " (" + str(c0 - ci[3]), str(c0 + ci[3]) + ")")
            c1 = self.metrics[species_id]["incidence_c1"][-1]
        #    print("C0: " + str(c1) + " (" + str(c1 - ci[4]), str(c1 + ci[4]) + ")")
        #    print()
        return

    def apply(self, data: pd.DataFrame | EventLog | Trace, verbose=True) -> None:
        """
        add all observations of an event log and update diversity and completeness profiles once afterward.
        If parameter step_size is set to an int, profiles are additionally updated along the way according to
        the step size
        :param data: the event log containing the trace observations
        """
        if isinstance(data, pd.DataFrame):
            return self.apply(pm4py.convert_to_event_log(data))

        #todo find out why this is notably faster than self.apply(tr) for tr in Log
        elif isinstance(data, EventLog) or isinstance(data, list) :
            for species_id in self.species_retrieval.keys():
                #TODO find better way to
                for tr in tqdm(data, "Profiling Log for " + species_id, disable=not verbose):
                    self.add_observation(tr, species_id)
                    # if step size is set, update metrics after <step_size> many traces
                    if self.step_size is None:
                        continue
                    elif self.metrics[species_id].incidence_sample_size % self.step_size == 0:
                        self.update_metrics(species_id)
                if self.step_size is None or len(data) % self.step_size != 0:
                    self.update_metrics(species_id)
            if self.no_bootstrap_samples > 0:
                self.add_bootstrap_ci(self.no_bootstrap_samples)
        elif isinstance(data, Trace):
            for species_id in self.species_retrieval.keys():
                self.add_observation(data, species_id)
                # if step size is set, update metrics after <step_size> many traces
                if self.step_size is None:
                    continue
                elif self.metrics[species_id].incidence_sample_size % self.step_size == 0:
                    self.update_metrics(species_id)

        else:
            raise RuntimeError('Cannot apply data of type ' + str(type(data)))

    def add_observation(self, observation: Trace, species_id: str) -> None:
        """
        adds a single observation
        :param species_id: the species definition for which observation shall be added
        :param observation: the trace observation
        """
        # retrieve species from current observation
        species_abundance = self.species_retrieval[species_id](observation)
        species_incidence = set(species_abundance)
        if len(species_abundance) == 0:
            self.metrics[species_id].empty_traces = self.metrics[species_id].empty_traces + 1
            self.current_obs_empty = True
        else:
            self.current_obs_empty = False

        self.metrics[species_id].trace_retrieved_species_abundance = species_abundance
        self.metrics[species_id].trace_retrieved_species_incidence = species_incidence

        # update species abundances/incidences
        for s in species_abundance:
            self.metrics[species_id].reference_sample_abundance[s] = self.metrics[
                                                                         species_id].reference_sample_abundance.get(s,
                                                                                                                    0) + 1

        for s in species_incidence:
            self.metrics[species_id].reference_sample_incidence[s] = self.metrics[
                                                                         species_id].reference_sample_incidence.get(s,
                                                                                                                    0) + 1

        # update current number of observation for each model
        self.metrics[species_id].abundance_sample_size = self.metrics[species_id].abundance_sample_size + len(
            species_abundance)
        self.metrics[species_id].incidence_sample_size = self.metrics[species_id].incidence_sample_size + 1

        # update current sum of all observed species for each model
        self.metrics[species_id].abundance_current_total_species_count = \
            self.metrics[species_id].abundance_current_total_species_count + len(species_abundance)
        self.metrics[species_id].incidence_current_total_species_count = \
            self.metrics[species_id].incidence_current_total_species_count + len(
                species_incidence)

        #update current degree of spatial aggregation
        if self.metrics[species_id].incidence_current_total_species_count == 0:
            self.metrics[species_id].current_co_occurrence = 0
        else:
            self.metrics[species_id].current_co_occurrence = 1 - (
                    self.metrics[species_id].incidence_current_total_species_count / self.metrics[
                species_id].abundance_current_total_species_count)

    def update_metrics(self, species_id: str) -> None:
        """
        updates the diversity and completeness profiles based on the current observations
        """
        #if self.current_obs_empty:
        #    return

        # update number of observations so far
        self.metrics[species_id]["abundance_no_observations"].append(self.metrics[species_id].abundance_sample_size)
        self.metrics[species_id]["incidence_no_observations"].append(self.metrics[species_id].incidence_sample_size)

        #update number of species seen so far
        self.metrics[species_id]["abundance_sum_species_counts"].append(
            self.metrics[species_id].abundance_current_total_species_count)
        self.metrics[species_id]["incidence_sum_species_counts"].append(
            self.metrics[species_id].incidence_current_total_species_count)

        #update degree of spatial aggregation
        self.metrics[species_id]["degree_of_co_occurrence"].append(self.metrics[species_id].current_co_occurrence)

        #update singleton and doubleton counts
        self.metrics[species_id]["abundance_singletons"].append(
            get_singletons(self.metrics[species_id].reference_sample_abundance))
        self.metrics[species_id]["incidence_singletons"].append(
            get_singletons(self.metrics[species_id].reference_sample_incidence))

        self.metrics[species_id]["abundance_doubletons"].append(
            get_doubletons(self.metrics[species_id].reference_sample_abundance))
        self.metrics[species_id]["incidence_doubletons"].append(
            get_doubletons(self.metrics[species_id].reference_sample_incidence))

        #update diversity profile
        if self.include_d0:
            self.__update_d0(species_id)
        if self.include_d1:
            self.__update_d1(species_id)
        if self.include_d2:
            self.__update_d2(species_id)

        #update completeness profile
        if self.include_c0:
            self.__update_c0(species_id)
        if self.include_c1:
            self.__update_c1(species_id)

        #update estimated sampling effort for target completeness
        for l in self.l_n:
            self.__update_l(l, species_id)

    def __update_d0(self, species_id: str) -> None:
        """
        updates D0 (=species richness) based on the current observations
        """
        #update sample metrics
        self.metrics[species_id]["abundance_sample_d0"].append(len(self.metrics[species_id].reference_sample_abundance))
        self.metrics[species_id]["incidence_sample_d0"].append(len(self.metrics[species_id].reference_sample_incidence))

        #update estimated metrics
        self.metrics[species_id]["abundance_estimate_d0"].append(
            hill_number_asymptotic(0, self.metrics[species_id].reference_sample_abundance,
                                   self.metrics[species_id].abundance_sample_size))
        self.metrics[species_id]["incidence_estimate_d0"].append(
            hill_number_asymptotic(0, self.metrics[species_id].reference_sample_incidence,
                                   self.metrics[species_id].incidence_sample_size, abundance=False))

        self.metrics[species_id]["incidence_estimate_d0_ci"].append(-1)

    def __update_d1(self, species_id: str) -> None:
        """
        updates D1 (=exponential of Shannon entropy) based on the current observations
        """
        #update sample metrics
        self.metrics[species_id]["abundance_sample_d1"].append(
            entropy_exp(self.metrics[species_id].reference_sample_abundance))
        self.metrics[species_id]["incidence_sample_d1"].append(
            entropy_exp(self.metrics[species_id].reference_sample_incidence))

        #update estimated metrics
        self.metrics[species_id]["abundance_estimate_d1"].append(
            hill_number_asymptotic(1, self.metrics[species_id].reference_sample_abundance,
                                   self.metrics[species_id].abundance_sample_size))
        self.metrics[species_id]["incidence_estimate_d1"].append(
            hill_number_asymptotic(1, self.metrics[species_id].reference_sample_incidence,
                                   self.metrics[species_id].incidence_sample_size, abundance=False))

        self.metrics[species_id]["incidence_estimate_d1_ci"].append(-1)


    def __update_d2(self, species_id: str) -> None:
        """
        updates D2 (=Simpson Diversity Index) based on the current observations
        """
        #update sample metrics
        self.metrics[species_id]["abundance_sample_d2"].append(
            simpson_diversity(self.metrics[species_id].reference_sample_abundance))
        self.metrics[species_id]["incidence_sample_d2"].append(
            simpson_diversity(self.metrics[species_id].reference_sample_incidence))

        #update estimated metrics
        self.metrics[species_id]["abundance_estimate_d2"].append(
            hill_number_asymptotic(2, self.metrics[species_id].reference_sample_abundance,
                                   self.metrics[species_id].abundance_sample_size))
        self.metrics[species_id]["incidence_estimate_d2"].append(
            hill_number_asymptotic(2, self.metrics[species_id].reference_sample_incidence,
                                   self.metrics[species_id].incidence_sample_size, abundance=False))

        self.metrics[species_id]["incidence_estimate_d2_ci"].append(-1)


    def __update_c0(self, species_id: str) -> None:
        """
        updates C0 (=completeness) based on the current observations
        """
        self.metrics[species_id]["abundance_c0"].append(
            completeness(self.metrics[species_id].reference_sample_abundance))
        self.metrics[species_id]["incidence_c0"].append(
            completeness(self.metrics[species_id].reference_sample_incidence))

        self.metrics[species_id]["incidence_c0_ci"].append(-1)

    def __update_c1(self, species_id: str) -> None:
        """
        updates C1 (=coverage) based on the current observations
        """
        self.metrics[species_id]["abundance_c1"].append(
            coverage(self.metrics[species_id].reference_sample_abundance,
                     self.metrics[species_id].abundance_sample_size))
        self.metrics[species_id]["incidence_c1"].append(
            coverage(self.metrics[species_id].reference_sample_incidence,
                     self.metrics[species_id].incidence_sample_size))

        self.metrics[species_id]["incidence_c1_ci"].append(-1)

    def __update_l(self, g: float, species_id: str) -> None:
        """
        updates l_g (=expected number additional observations for reaching completeness g) based on the current
        observations
        :param g: desired  completeness
        """
        self.metrics[species_id]["abundance_l_" + str(g)].append(
            sampling_effort_abundance(g, self.metrics[species_id].reference_sample_abundance,
                                      self.metrics[species_id].abundance_sample_size))
        self.metrics[species_id]["incidence_l_" + str(g)].append(
            sampling_effort_incidence(g, self.metrics[species_id].reference_sample_incidence,
                                      self.metrics[species_id].incidence_sample_size))

    def summarize(self, species_id: str = None) -> None:
        """
        prints a summary of the species profiles of the current reference sample.

        """
        species_ids = self.metrics.keys() if species_id is None else [species_id]

        for species_id in species_ids:
            print("### "+species_id+" ###")
            print("%-25s %-20s %s" % ("Sample Stats", "Abundance", "Incidence"))
            print("%-25s %-20s %s" % ("", "---------", "---------"))
            print("%-25s %-20s %s" % ("No Observations", str(self.metrics[species_id]["abundance_no_observations"][-1]),str(self.metrics[species_id]["incidence_no_observations"][-1])))
            print("%-25s %-20s %s" % ("No Species", str(self.metrics[species_id]["abundance_sum_species_counts"][-1]),str(self.metrics[species_id]["incidence_sum_species_counts"][-1])))

            print("%-25s %-20s %s" % ("Singletons", str(self.metrics[species_id]["abundance_singletons"][-1]),str(self.metrics[species_id]["abundance_doubletons"][-1])))
            print("%-25s %-20s %s" % ("Doubletons", str(self.metrics[species_id]["abundance_doubletons"][-1]),str(self.metrics[species_id]["incidence_doubletons"][-1])))
            print("%-25s %s" % ("Degree of Co-Occurrence", str(self.metrics[species_id]["degree_of_co_occurrence"][-1])))
            print()
            print("%-25s %-20s %-20s %s" % ("Abundance:", "Observed", "Estimate", "Stdev"))
            print("%-25s %-20s %-20s %s" % ("", "--------", "--------", "-----"))
            print("%-25s %-20s %-20s %s" % ("D0",str(self.metrics[species_id]["abundance_sample_d0"][-1]), str(self.metrics[species_id]["abundance_estimate_d0"][-1]),"-"))
            print("%-25s %-20s %-20s %s" % ("D1",str(self.metrics[species_id]["abundance_sample_d1"][-1]), str(self.metrics[species_id]["abundance_estimate_d1"][-1]),"-"))
            print("%-25s %-20s %-20s %s" % ("D2",str(self.metrics[species_id]["abundance_sample_d2"][-1]), str(self.metrics[species_id]["abundance_estimate_d2"][-1]),"-"))
            print("%-25s %-20s %-20s %s" % ("C0","-", str(self.metrics[species_id]["abundance_c0"][-1]),"-"))
            print("%-25s %-20s %-20s %s" % ("C1","-", str(self.metrics[species_id]["abundance_c1"][-1]),"-"))
            for l in self.l_n:
                print("%-25s %-20s %-20s %s" %("l_"+str(l), "-", str(self.metrics[species_id]["abundance_l_" + str(l)][-1]), "-"))
            print()
            print("%-25s %-20s %-20s %s" % ("Incidence:", "Observed", "Estimate", "Stdev"))
            print("%-25s %-20s %-20s %s" % ("", "--------", "--------", "-----"))
            print("%-25s %-20s %-20s %s" % ("D0",str(self.metrics[species_id]["incidence_sample_d0"][-1]), str(self.metrics[species_id]["incidence_estimate_d0"][-1]),str(self.metrics[species_id]["incidence_estimate_d0_ci"][-1]) if self.no_bootstrap_samples>0 else "-"))
            print("%-25s %-20s %-20s %s" % ("D1",str(self.metrics[species_id]["incidence_sample_d1"][-1]), str(self.metrics[species_id]["incidence_estimate_d1"][-1]),str(self.metrics[species_id]["incidence_estimate_d1_ci"][-1]) if self.no_bootstrap_samples>0 else "-"))
            print("%-25s %-20s %-20s %s" % ("D2",str(self.metrics[species_id]["incidence_sample_d2"][-1]), str(self.metrics[species_id]["incidence_estimate_d2"][-1]),str(self.metrics[species_id]["incidence_estimate_d2_ci"][-1]) if self.no_bootstrap_samples>0 else "-"))
            print("%-25s %-20s %-20s %s" % ("C0","-", str(self.metrics[species_id]["incidence_c0"][-1]),str(self.metrics[species_id]["incidence_c0_ci"][-1]) if self.no_bootstrap_samples>0 else "-"))
            print("%-25s %-20s %-20s %s" % ("C1","-", str(self.metrics[species_id]["incidence_c1"][-1]),str(self.metrics[species_id]["incidence_c1_ci"][-1]) if self.no_bootstrap_samples>0 else "-"))
            for l in self.l_n:
                print("%-25s %-20s %-20s %s" %("l_"+str(l), "-", str(self.metrics[species_id]["incidence_l_" + str(l)][-1]),"-"))
            print("")
            print("")


    @deprecated
    def print_metrics(self) -> None:
        """
        prints the Diversity and Completeness Profile of the current observations
        """
        for species_id in self.species_retrieval:
            print("### " + species_id + " ###")
            print("### SAMPLE STATS ###")
            print("Abundance")
            print("%-30s %s" % ("     No Observations:", str(self.metrics[species_id]["abundance_no_observations"])))
            print("%-30s %s" % (
                "     Total Species Count:", str(self.metrics[species_id]["abundance_sum_species_counts"])))
            print("%-30s %s" % ("     Singletons:", str(self.metrics[species_id]["abundance_singletons"])))
            print("%-30s %s" % ("     Doubletons:", str(self.metrics[species_id]["abundance_doubletons"])))

            print("Incidence")
            print("%-30s %s" % ("     No Observations:", str(self.metrics[species_id]["incidence_no_observations"])))
            print("%-30s %s" % (
                "     Total Species Count:", str(self.metrics[species_id]["incidence_sum_species_counts"])))
            print("%-30s %s" % ("     Singletons:", str(self.metrics[species_id]["incidence_singletons"])))
            print("%-30s %s" % ("     Doubletons:", str(self.metrics[species_id]["incidence_doubletons"])))

            print("%-30s %s" % ("     Empty Traces:", str(self.metrics[species_id].empty_traces)))
            print("%-30s %s" % ("Degree of Co-Occurrence:", str(self.metrics[species_id]["degree_of_aggregation"])))
            print()
            print("### DIVERSITY AND COMPLETENESS PROFILE ###")
            print("Abundance")
            if self.include_d0:
                print("%-30s %s" % ("     D0 - sample:", (self.metrics[species_id]["abundance_sample_d0"])))
                print("%-30s %s" % ("     D0 - estimate:", str(self.metrics[species_id]["abundance_estimate_d0"])))
            if self.include_d1:
                print("%-30s %s" % ("     D1 - sample:", str(self.metrics[species_id]["abundance_sample_d1"])))
                print("%-30s %s" % ("     D1 - estimate:", str(self.metrics[species_id]["abundance_estimate_d1"])))
            if self.include_d2:
                print("%-30s %s" % ("     D2 - sample:", str(self.metrics[species_id]["abundance_sample_d2"])))
                print("%-30s %s" % ("     D2 - estimate:", str(self.metrics[species_id]["abundance_estimate_d2"])))
            if self.include_c0:
                print("%-30s %s" % ("     C0:", str(self.metrics[species_id]["abundance_c0"])))
            if self.include_c1:
                print("%-30s %s" % ("     C1:", str(self.metrics[species_id]["abundance_c1"])))
            for l in self.l_n:
                print("%-30s %s" % ("     l_" + str(l) + ":", str(self.metrics[species_id]["abundance_l_" + str(l)])))
            print("Incidence")
            if self.include_d0:
                print("%-30s %s" % ("     D0 - sample:", (self.metrics[species_id]["incidence_sample_d0"])))
                print("%-30s %s" % ("     D0 - estimate:", str(self.metrics[species_id]["incidence_estimate_d0"])))
                print("%-30s %s" % ("     D0 - CI:", str(self.metrics[species_id]["incidence_estimate_d0_ci"])))

            if self.include_d1:
                print("%-30s %s" % ("     D1 - sample:", str(self.metrics[species_id]["incidence_sample_d1"])))
                print("%-30s %s" % ("     D1 - estimate:", str(self.metrics[species_id]["incidence_estimate_d1"])))
                print("%-30s %s" % ("     D1 - CI:", str(self.metrics[species_id]["incidence_estimate_d1_ci"])))

            if self.include_d2:
                print("%-30s %s" % ("     D2 - sample:", str(self.metrics[species_id]["incidence_sample_d2"])))
                print("%-30s %s" % ("     D2 - estimate:", str(self.metrics[species_id]["incidence_estimate_d2"])))
                print("%-30s %s" % ("     D2 - CI:", str(self.metrics[species_id]["incidence_estimate_d2_ci"])))

            if self.include_c0:
                print("%-30s %s" % ("     C0:", str(self.metrics[species_id]["incidence_c0"])))
                print("%-30s %s" % ("     C0 - CI:", str(self.metrics[species_id]["incidence_c0_ci"])))

            if self.include_c1:
                print("%-30s %s" % ("     C1:", str(self.metrics[species_id]["incidence_c1"])))
                print("%-30s %s" % ("     C1 - CI:", str(self.metrics[species_id]["incidence_c1_ci"])))

            for l in self.l_n:
                print("%-30s %s" % ("     l_" + str(l) + ":", str(self.metrics[species_id]["incidence_l_" + str(l)])))
            print()

    def to_dataFrame(self, include_all=True) -> DataFrame:
        """
        returns the diversity and completeness profile of the current observations as a data frame
        :returns: a data frame view of the Diversity and Completeness Profile
        """
        return pd.DataFrame([[i, j, ix, v]
                             for i in self.metrics.keys()
                             for j in self.metrics[i].keys()
                             for ix, v in enumerate(self.metrics[i][j])
                             ], columns=["species", "metric", "observation", "value"]
                            ) if include_all else pd.DataFrame([[i, j, self.metrics[i][j][-1]]
                                                                for i in self.metrics.keys()
                                                                for j in self.metrics[i].keys()
                                                                ], columns=["species", "metric", "value"])



import math
from functools import cache
from src.estimation.metrics import get_incidence_count
import gmpy2


def rarefy_richness_abundance(reference_sample, sample_size, goal, data_points=100):
    m_size = math.floor(sample_size / data_points)
    s_obs = len(reference_sample)
    values = []
    locations = []
    m = 0
    while m + m_size < sample_size:
        print("Richness Rarefaction - m=" + str(m))
        m = m + m_size
        s = 0
        for y_i in reference_sample.values():
            first = binom(sample_size - y_i, m)  # binom(sample_size - y_i, m)
            second = binom(sample_size, m)  # binom(sample_size, m)
            s = s + (first / second)
        values.append(s_obs - s)
        locations.append(m)
    values.insert(0, 0)
    locations.insert(0, 0)
    return values, locations


# structurally equal to abundance-based case
def rarefy_richness_incidence(reference_sample, sample_size, goal, data_points=100):
    return rarefy_richness_abundance(reference_sample, sample_size, goal, data_points)


def rarefy_shannon_entropy_abundance(reference_sample, sample_size, goal, data_points=100):
    incidences = []
    for x in set(reference_sample.values()):
        incidences.append((x, get_incidence_count(reference_sample, x)))
    incidences = tuple(incidences)

    # incidences[x] = get_incidence_count(reference_sample,x)
    m_size = math.floor(sample_size / data_points)
    values = []
    locations = []
    m = 0
    while m + m_size <= sample_size:
        print("Shannon Entropy Rarefaction - m=" + str(m))
        s = 0
        m = m + m_size

        for k in range(1, m + 1):
            first = (-(k / m) * math.log(k / m))
            second = get_expected_incidence_of_k_at_m(sample_size, k, m, incidences)
            v = first * second
            s = s + v
        values.append(math.exp(s))
        locations.append(m)
    values.insert(0, 0)
    locations.insert(0, 0)
    return values, locations


def rarefy_shannon_entropy_incidence(reference_sample, sample_size, goal, data_points=100):
    incidences = []
    for x in set(reference_sample.values()):
        incidences.append((x, get_incidence_count(reference_sample, x)))
    incidences = tuple(incidences)

    m_size = math.floor(sample_size / data_points)
    u = sum(list(reference_sample.values()))
    values = []
    locations = []
    m = 0
    while m + m_size <= sample_size:
        s = 0
        m = m + m_size
        print("Shannon Entropy Rarefaction - m=" + str(m))
        u_t = (m * u) / sample_size
        for k in range(1, m + 1):
            first = (-(k / u_t) * math.log(k / u_t))
            second = get_expected_incidence_of_k_at_m(sample_size, k, m, incidences)
            v = first * second
            # print(first, second)
            s = s + v
        values.append(math.exp(s))
        locations.append(m)
    values.insert(0, 0)
    locations.insert(0, 0)
    return values, locations


def rarefy_simpson_diversity_abundance(reference_sample, sample_size, goal, data_points=100):
    incidences = []
    for x in set(reference_sample.values()):
        incidences.append((x, get_incidence_count(reference_sample, x)))
    incidences = tuple(incidences)

    m_size = math.floor(sample_size / data_points)
    # s_obs = len(reference_sample)
    # u = sum(list(reference_sample.values()))
    values = []
    locations = []
    m = 0
    while m + m_size < sample_size:
        print("Simpson Diversity Rarefaction - m=" + str(m))
        m = m + m_size
        # u_t = m * u / sample_size
        value = 1 / (
            sum([(k / m) ** 2 * get_expected_incidence_of_k_at_m(sample_size, k, m, incidences) for k in
                 range(1, m + 1)]))
        values.append(value)
        locations.append(m)
    values.insert(0, 0)
    locations.insert(0, 0)
    return values, locations


def rarefy_simpson_diversity_incidence(reference_sample, sample_size, goal, data_points=100):
    incidences = []
    for x in set(reference_sample.values()):
        incidences.append((x, get_incidence_count(reference_sample, x)))
    incidences = tuple(incidences)
    m_size = math.floor(sample_size / data_points)
    u = sum(list(reference_sample.values()))
    values = []
    locations = []
    m = 0
    while m + m_size < sample_size:
        print("Simpson Diversity Rarefaction - m=" + str(m))
        m = m + m_size
        u_t = m * u / sample_size
        value = 1 / (
            sum([(k / u_t) ** 2 * get_expected_incidence_of_k_at_m(sample_size, k, m, incidences) for k in
                 range(1, m + 1)]))
        values.append(value)
        locations.append(m)
    values.insert(0, 0)
    locations.insert(0, 0)
    return values, locations


@cache
def get_expected_incidence_of_k_at_m(sample_size, k, m, incidences):
    s = 0
    # for j in set(reference_sample.values()):
    for j, f_j in incidences:
        # for j in range(k, sample_size):
        if j < k:
            continue
        # f_j = get_incidence_count(reference_sample, j)
        # if sample_size-j < m-k:
        #   continue
        # if f_j == 0:# or f_j > sample_size - m+k+1:
        #     continue

        # first = scipy.special4pm.comb(j, k, exact=True) if k <= j / 2 else scipy.special4pm.comb(j, j - k, exact=True)
        # second = scipy.special4pm.comb(sample_size - j, m - k, exact=True) if m - k <= (
        #        sample_size - j) / 2 else scipy.special4pm.comb(sample_size - j, sample_size - j - m - k, exact=True)
        # third = scipy.special4pm.comb(sample_size, m, exact=True) if m <= sample_size else scipy.special4pm.comb(sample_size,
        #                                                                                                   sample_size - m,
        #                                                                                                   exact=True)
        first = binom(j, k)  # if k <= j / 2 else binom(j, j - k)
        second = binom(sample_size - j,
                       m - k)  # if m - k <= (sample_size - j) / 2 else binom(sample_size - j, sample_size - j - m - k)
        third = binom(sample_size, m)  # if m <= sample_size else binom(sample_size,sample_size - m)
        # print(first, second, third)
        # s = s * f_j * math.exp(first + second - third)
        s = s + ((first * second) / third) * f_j
    return s

    # first, second, third = 0, 0, 0
    # s = 0
    # third = scipy.special4pm.comb(sample_size, m, exact=True) if m<=sample_size else scipy.special4pm.comb(sample_size, sample_size - m, exact=True)
    # # print()
    # for i, y_i in enumerate(reference_sample.values()):
    # #    # print(y_i)
    #     if y_i >= k:
    #         first = scipy.special4pm.comb(y_i,k, exact=True) if k<=y_i/2 else scipy.special4pm.comb(y_i,y_i-k, exact=True)
    #         if sample_size - y_i >= m - k:
    #             second = scipy.special4pm.comb(sample_size-y_i, m-k, exact=True) if m-k<=(sample_size-y_i)/2 else scipy.special4pm.comb(sample_size-y_i, sample_size-y_i - m-k, exact=True)
    # #            # print(first, second, first * second, third)
    # #    # print(first * second, third)
    #     s = s + (first * second) / (third)
    #
    # return s


@cache
def binom(n, k):
    # if k>n:
    #    return 0
    # return math.comb(n, k)
    # return scipy.special4pm.comb(n, k, exact=True)
    return gmpy2.comb(n, k)
    # """
    # A fast way to calculate binomial coefficients by Andrew Dalke (contrib).
    # """
    # if 0 <= k <= n:
    #     ntok = 1
    #     ktok = 1
    #     for t in range(1, min(k, n - k) + 1):
    #         ntok *= n
    #         ktok *= t
    #         n -= 1
    #     return ntok // ktok
    # else:
    #     return 0
    # """Computes n! / (r! (n-r)!) exactly. Returns a python long int."""
    # assert n >= 0
    # assert 0 <= r <= n
    #
    # c = 1
    # denom = 1
    # for (num, denom) in zip(range(n, n - r, -1), range(1, r + 1, 1)):
    #     c = (c * num) // denom
    # return c


def binomial(n, k):
    if k > n:
        return 0
    if k > n / 2:
        k = n - k
    prod = 1
    for i in range(1, k + 1):
        prod = prod * (n + 1 - i) / i
    # print("!!!!------------!!!!")
    return prod
    # return reduce(operator.mul, [(n+1-i)/i for i in range(1,k+1)], 1)

    # (math.prod([(n+1-i)/i for i in range(1,k+1)]))



import math
import statistics
import time


from src.bootstrap.bootstrap import generate_bootstrap_sequence_abundance, generate_bootstrap_sequence_incidence
from src.raripolation.extrapolation import extrapolate_richness_abundance, extrapolate_shannon_entropy_abundance, \
    extrapolate_simpson_diversity_abundance, extrapolate_richness_incidence, extrapolate_shannon_entropy_incidence, \
    extrapolate_simpson_diversity_incidence
from src.estimation.metrics import get_number_observed_species, entropy_exp, simpson_diversity, estimate_species_richness_chao, estimate_shannon_entropy_abundance, \
    estimate_shannon_entropy_incidence
from rarefaction import rarefy_richness_abundance, rarefy_shannon_entropy_abundance, rarefy_simpson_diversity_abundance, \
    rarefy_richness_incidence, rarefy_shannon_entropy_incidence, rarefy_simpson_diversity_incidence


def rarefy_extrapolate_q0(est, abundance=True, data_points=30):
    start = time.time()
    if abundance:
        print("Richness Abundance(" + str(est.observed_species_count[-1])+", "+ str(est.chao2_abundance[-1]) + ")")

        r, loc_r = rarefy_richness_abundance(est.reference_sample_abundance, est.number_observations_abundance,
                                             est.observed_species_count[-1], data_points)
        e, loc_e = extrapolate_richness_abundance(est.reference_sample_abundance, est.number_observations_abundance,
                                                  est.chao2_abundance[-1], data_points)
        print("Eplapsed Time: "+str(time.time() - start))
        return r.append + [est.observed_species_count[-1]] + e, loc_r + [est.number_observations_abundance] + loc_e
    else:
        print("Richness Incidence(" + str(est.observed_species_count[-1])+", "+ str(est.chao2_incidence[-1]) + ")")

        r, loc_r = rarefy_richness_incidence(est.reference_sample_incidence, est.number_observations_incidence,
                                             est.observed_species_count[-1], data_points)
        e, loc_e = extrapolate_richness_incidence(est.reference_sample_incidence, est.number_observations_incidence,
                                                  est.chao2_abundance[-1], data_points)
        print("Eplapsed Time: "+str(time.time() - start))
        return r + [est.observed_species_count[-1]] + e, loc_r + [est.number_observations_incidence] + loc_e


def rarefy_extrapolate_q1(est, abundance=True, data_points=30):
    start = time.time()
    if abundance:
        print("Shannon Abundance(" + str(est.D1_sample_abundance[-1])+", "+ str(est.D1_estimated_abundance[-1]) + ")")

        r, loc_r = rarefy_shannon_entropy_abundance(est.reference_sample_abundance, est.number_observations_abundance,
                                                    est.D1_sample_abundance[-1], data_points)
        e, loc_e = extrapolate_shannon_entropy_abundance(est.reference_sample_abundance,
                                                         est.number_observations_abundance,
                                                         est.D1_estimated_abundance[-1], data_points)
        print("Eplapsed Time: "+str(time.time() - start))
        return r + [est.D1_sample_abundance[-1]] + e, loc_r + [est.number_observations_abundance] + loc_e
    else:
        print("Shannon Incidence(" + str(est.D1_sample_incidence[-1])+", "+ str(est.D1_estimated_incidence[-1]) + ")")

        r, loc_r = rarefy_shannon_entropy_incidence(est.reference_sample_incidence, est.number_observations_incidence,
                                                    est.D1_sample_incidence[-1], data_points)
        e, loc_e = extrapolate_shannon_entropy_incidence(est.reference_sample_incidence,
                                                         est.number_observations_incidence,
                                                         est.D1_estimated_incidence[-1], data_points)
        print("Eplapsed Time: "+str(time.time() - start))
        return r + [est.D1_sample_incidence[-1]] + e, loc_r + [est.number_observations_incidence] + loc_e


def rarefy_extrapolate_q2(est, abundance=True, data_points=30):
    start = time.time()
    if abundance:
        print("Simpson Abundance(" + str(est.D2_sample_abundance[-1])+", "+ str(est.D2_estimated_abundance[-1]) + ")")

        r, loc_r = rarefy_simpson_diversity_abundance(est.reference_sample_abundance, est.number_observations_abundance,
                                                      est.D2_sample_abundance[-1], data_points)
        e, loc_e = extrapolate_simpson_diversity_abundance(est.reference_sample_abundance,
                                                           est.number_observations_abundance,
                                                           data_points)
        print("Eplapsed Time: "+str(time.time() - start))
        return r + [est.D2_sample_abundance[-1]] + e, loc_r + [est.number_observations_abundance] + loc_e
    else:
        print("Simpson Incidence(" + str(est.D2_sample_incidence[-1])+", "+ str(est.D2_estimated_incidence[-1]) + ")")
        r, loc_r = rarefy_simpson_diversity_incidence(est.reference_sample_incidence, est.number_observations_incidence,
                                                      est.D2_sample_incidence[-1], data_points)
        e, loc_e = extrapolate_simpson_diversity_incidence(est.reference_sample_incidence,
                                                           est.number_observations_incidence,
                                                           data_points)
        print("Eplapsed Time: "+str(time.time() - start))
        return r + [est.D2_sample_incidence[-1]] + e, loc_r + [est.number_observations_incidence] + loc_e


def rarefy_extrapolate_all(est, abundance_data=True, data_points=30):
    if abundance_data:
        q0_a, q0_a_loc = rarefy_extrapolate_q0(est, abundance=True, data_points=data_points)
        q1_a, q1_a_loc = rarefy_extrapolate_q1(est, abundance=True, data_points=data_points)
        q2_a, q2_a_loc = rarefy_extrapolate_q2(est, abundance=True, data_points=data_points)
        return (q0_a, q0_a_loc), (q1_a, q1_a_loc), (q2_a, q2_a_loc)
    else:
        q0_i, q0_i_loc = rarefy_extrapolate_q0(est, abundance=False, data_points=data_points)
        q1_i, q1_i_loc = rarefy_extrapolate_q1(est, abundance=False, data_points=data_points)
        q2_i, q2_i_loc = rarefy_extrapolate_q2(est, abundance=False, data_points=data_points)
        return (q0_i, q0_i_loc), (q1_i, q1_i_loc), (q2_i, q2_i_loc)



def rarefy_extrapolate_bootstrap_all(s, abundance_data=True, data_points=30, bootstrap_repetitions = 200):
    q0_all=[]
    q1_all=[]
    q2_all=[]

    q_loc=None
    print("Creatign Bootstrap Samples")
    for i in range(0, bootstrap_repetitions):
        #function results for all samples
        q0 = [0]
        q1 = [0]
        q2 = [0]

        seq_a = None
        loc_a = None
        #subsampling of bootstrap samples
        if abundance_data:
            seq_a, loc_a = generate_bootstrap_sequence_abundance(s.reference_sample_abundance,
                                                                 s.number_observations_abundance, math.floor(
                    s.number_observations_abundance / data_points))
        else:
            #print("OBSERVATIONS: "+str(s.number_observations_incidence))
            seq_a, loc_a = generate_bootstrap_sequence_incidence(s.reference_sample_incidence,
                                                                 s.number_observations_incidence, math.floor(
                    s.number_observations_incidence / data_points))


        for seq in seq_a:
            q0.append(get_number_observed_species(seq))
            q1.append(entropy_exp(seq,sum(seq.values())))
            q2.append(simpson_diversity(seq,sum(seq.values())))

        #extrapolation
        ref_sample = seq_a[-1]
        q0_e = None
        q1_e = None
        q2_e = None
        loc_e = None

        if abundance_data:
            q0_e, loc_e = extrapolate_richness_abundance(ref_sample, s.number_observations_abundance, estimate_species_richness_chao(ref_sample), data_points=data_points)
            q1_e, _ = extrapolate_shannon_entropy_abundance(ref_sample, s.number_observations_abundance,
                                                            math.exp(estimate_shannon_entropy_abundance(ref_sample, s.number_observations_abundance)), data_points= data_points)
            q2_e, _ = extrapolate_simpson_diversity_abundance(ref_sample, s.number_observations_abundance,data_points= data_points)
        else:
            q0_e, loc_e = extrapolate_richness_incidence(ref_sample, s.number_observations_incidence, estimate_species_richness_chao(ref_sample),
                                                         data_points=data_points)
            q1_e, _ = extrapolate_shannon_entropy_incidence(ref_sample, s.number_observations_incidence,
                                                            math.exp(estimate_shannon_entropy_incidence(ref_sample,
                                                                                                        s.number_observations_incidence)),
                                                            data_points=data_points)
            q2_e, _ = extrapolate_simpson_diversity_incidence(ref_sample, s.number_observations_incidence,
                                                              data_points=data_points)

        #connect subsampled and extrapolated function curves
        q0 = q0 + q0_e
        q1 = q1 + q1_e
        q2 = q2 + q2_e
        q_loc = loc_a + loc_e

        #add bootstrap function values to list of all results
        for x in range(0,len(q_loc)):
            if len(q0_all)<=x:
                q0_all.append([])
                q1_all.append([])
                q2_all.append([])
            q0_all[x].append(q0[x])
            q1_all[x].append(q1[x])
            q2_all[x].append(q2[x])

    #aggregate bootstrap results
    q0_mean = []
    q0_ciu = []
    q0_cil = []

    q1_mean = []
    q1_ciu = []
    q1_cil = []

    q2_mean = []
    q2_ciu = []
    q2_cil = []
    for x in range(0,len(q0_all)):
        q0_mean.append(statistics.mean(q0_all[x]))
        q0_ciu.append(q0_mean[x]+1.96*statistics.stdev(q0_all[x]))
        q0_cil.append(q0_mean[x]-1.96*statistics.stdev(q0_all[x]))

        q1_mean.append(statistics.mean(q1_all[x]))
        q1_ciu.append(q1_mean[x]+1.96*statistics.stdev(q1_all[x]))
        q1_cil.append(q1_mean[x]-1.96*statistics.stdev(q1_all[x]))

        q2_mean.append(statistics.mean(q2_all[x]))
        q2_ciu.append(q2_mean[x]+1.96*statistics.stdev(q2_all[x]))
        q2_cil.append(q2_mean[x]-1.96*statistics.stdev(q2_all[x]))

    return(q0_mean,q1_mean,q2_mean),(q0_cil,q1_cil,q2_cil),(q0_ciu,q1_ciu,q2_ciu), q_loc


import unittest

from process_completeness_estimation.estimation.metrics import completeness, coverage


class TestCompleteness(unittest.TestCase):
    def test_completeness_no_observations(self):
        self.assertEqual(completeness({}), 0.0)

    def test_completeness_complete_observations(self):
        self.assertEqual(completeness({"A": 10, "B": 5, "C": 2, "D": 2}), 1.0)

    def test_completeness_incomplete_observations(self):
        self.assertLess(completeness({"A": 10, "B": 5, "C": 2, "D": 1}), 1.0)
        self.assertGreater(completeness({"A": 10, "B": 5, "C": 2, "D": 1}), 0.0)

    def test_completeness_comparing_observations(self):
        self.assertLess(completeness({"A": 10, "B": 5, "C": 1, "D": 1}),
                        completeness({"A": 10, "B": 5, "C": 2, "D": 1}))


class TestCoverage(unittest.TestCase):
    def test_coverage_no_observations(self):
        self.assertEqual(coverage({}, 0), 0.0)

    def test_coverage_complete_observations_abundance(self):
        self.assertEqual(coverage({"A": 10, "B": 5, "C": 2, "D": 2}, 19), 1.0)

    def test_coverage_complete_observations_incidence(self):
        self.assertEqual(coverage({"A": 10, "B": 5, "C": 2, "D": 2}, 10), 1.0)

    def test_coverage_incomplete_observations_abundance(self):
        self.assertLess(coverage({"A": 10, "B": 5, "C": 2, "D": 1}, 19), 1.0)
        self.assertGreater(coverage({"A": 10, "B": 5, "C": 2, "D": 1}, 19), 0.0)

    def test_coverage_incomplete_observations_incidence(self):
        self.assertLess(coverage({"A": 10, "B": 5, "C": 2, "D": 1}, 10), 1.0)
        self.assertGreater(coverage({"A": 10, "B": 5, "C": 2, "D": 1}, 10), 0.0)

    def test_coverage_comparing_observations_abundance(self):
        self.assertLess(coverage({"A": 10, "B": 5, "C": 1, "D": 1}, 19),
                        coverage({"A": 10, "B": 5, "C": 2, "D": 1}, 19))

    def test_coverage_comparing_observations_incidence_same_sample_size(self):
        self.assertLess(coverage({"A": 10, "B": 5, "C": 1, "D": 1}, 10),
                        coverage({"A": 10, "B": 5, "C": 2, "D": 1}, 10))

    def test_coverage_comparing_observations_incidence_same_observations(self):
        self.assertLess(coverage({"A": 10, "B": 5, "C": 2, "D": 1}, 15),
                        coverage({"A": 10, "B": 5, "C": 2, "D": 1}, 10))


from functools import partial

import numpy as np
import pandas as pd
import pm4py
from matplotlib import pyplot as plt

import species_estimator
import species_retrieval
from rarefaction_extrapolation import rarefy_extrapolate_bootstrap_all

WIDTH = 9

# TODO CONFIDENCE DYNAMIC!!!!!!
def summarize_metric(metric, estimators):
    df_all = pd.DataFrame([getattr(x, metric) for x in estimators],
                          columns=[x for x in range(0, len(getattr(estimators[0], metric)))])
    df_stats = pd.concat([
        pd.DataFrame([df_all[x].min()] for x in df_all.columns),
        pd.DataFrame([df_all[x].nsmallest(2, keep='all').iloc[-1]] for x in df_all.columns),
        pd.DataFrame(df_all.median()),
        pd.DataFrame([df_all[x].nlargest(2, keep='all').iloc[-1]] for x in df_all.columns),
        pd.DataFrame([df_all[x].max()] for x in df_all.columns)]
        , axis=1)
    df_stats.columns = ["min", "lower_ci", "median", "upper_ci", "max"]
    return df_stats

def plot_hill_numbers_raripolated(estimation,abundance = True):
    if abundance:
        est, lci, uci, loc = rarefy_extrapolate_bootstrap_all(estimation, abundance_data = True, data_points=50,
                                                              bootstrap_repetitions=100)
        plt.fill_between(loc,lci[0],uci[0], alpha=0.4)
        plt.plot(loc,est[0])

        plt.fill_between(loc,lci[1],uci[1], alpha=0.4)
        plt.plot(loc,est[1])

        plt.fill_between(loc,lci[2],uci[2], alpha=0.4)
        plt.plot(loc,est[2])

        print(estimation.number_observations_incidence,estimation.number_observations_abundance)
        plt.plot(estimation.observation_ids_abundance,estimation.observed_species_count)
        plt.plot(estimation.observation_ids_abundance,estimation.D1_sample_abundance)
        plt.plot(estimation.observation_ids_abundance,estimation.D2_sample_abundance)
    else:
        est, lci, uci, loc = rarefy_extrapolate_bootstrap_all(estimation, abundance_data=False, data_points=50,
                                                              bootstrap_repetitions=100)
        print()
        print(estimation.observed_species_count)
        print(estimation.D1_sample_incidence)
        print(estimation.D2_sample_incidence)
        print(est[0])
        print(est[1])
        print(est[2])
        plt.fill_between(loc, lci[0], uci[0], alpha=0.4)
        plt.plot(loc, est[0])

        plt.fill_between(loc, lci[1], uci[1], alpha=0.4)
        plt.plot(loc, est[1])

        plt.fill_between(loc, lci[2], uci[2], alpha=0.4)
        plt.plot(loc, est[2])

        plt.plot(estimation.observation_ids, estimation.observed_species_count)
        plt.plot(estimation.observation_ids, estimation.D1_sample_incidence)
        plt.plot(estimation.observation_ids, estimation.D2_sample_incidence)
    plt.show()


def plot_rank_abundance(estimation, name):
    #plt.style.use('seaborn-v0_8-ticks')

    plt.rcParams['figure.figsize'] = [WIDTH, 3]
    plt.rcParams['xtick.labelsize'] = 20
    plt.rcParams['ytick.labelsize'] = 20

    no_species = len(estimation.reference_sample_abundance.values())

    plt.plot(sorted(list(estimation.reference_sample_abundance.values()), reverse=True))
    plt.fill_between(np.linspace(0, no_species, no_species, endpoint=False),
                     sorted(list(estimation.reference_sample_abundance.values()), reverse=True),
                     [0 for _ in range(no_species)], alpha=0.4)
    #plt.title("Rank Abundance Curve", fontsize=22)
    plt.xlabel("Species Rank", fontsize=24)
    plt.ylabel("Occurrences", fontsize=24)
    plt.xticks([0, no_species-1], [1, no_species])
    plt.yticks([0, max(estimation.reference_sample_abundance.values())],
               [0, max(estimation.reference_sample_abundance.values())])

    plt.tight_layout()
    plt.savefig("figures/" + name + "_curve_abundance.pdf", format="pdf")
    plt.close()
    #plt.show()

    plt.rcParams['figure.figsize'] = [WIDTH, 3]
    plt.rcParams['xtick.labelsize'] = 20
    plt.rcParams['ytick.labelsize'] = 20

    no_species = len(estimation.reference_sample_incidence.values())

    plt.plot(sorted(list(estimation.reference_sample_incidence.values()), reverse=True))
    plt.fill_between(np.linspace(0, no_species, no_species, endpoint=False),
                     sorted(list(estimation.reference_sample_incidence.values()), reverse=True),
                     [0 for _ in range(no_species)], alpha=0.4)
    #plt.title("Species Occurence (Incidence)", fontsize=22)
    plt.xlabel("Species Rank", fontsize=24)
    plt.ylabel("Occurrences", fontsize=24)
    plt.xticks([0, no_species-1], [1, no_species])
    plt.yticks([0, max(estimation.reference_sample_abundance.values()) ],
               [0, max(estimation.reference_sample_abundance.values()) ])

    plt.tight_layout()
    plt.savefig("figures/" + name + "_curve_incidence.pdf", format="pdf")
    plt.close()
    #plt.show()


# here we only show abundance plots - trends are equivalent for both sides
def plot_rank_abundances_multiple_estimators(names, estimations, est_id):
    names = list(names)

    plt.rcParams['figure.figsize'] = [2*WIDTH, 3]
    plt.rcParams['xtick.labelsize'] = 20
    plt.rcParams['ytick.labelsize'] = 20

    f, (ax1, ax2, ax3, ax4) = plt.subplots(nrows=1, ncols=4, sharey="row")

    no_species = len(estimations[0].reference_sample_abundance.values())
    ax1.plot(sorted(list(estimations[0].reference_sample_abundance.values()), reverse=True))
    ax1.fill_between(np.linspace(0, no_species, no_species, endpoint=False),
                     sorted(list(estimations[0].reference_sample_abundance.values()), reverse=True),
                     [0 for _ in range(no_species)], alpha=0.5)
    ax1.set_title(names[0], fontsize=24)
    ax1.set_xlabel("Species Rank", fontsize=20)
    ax1.set_ylabel("Occurences", fontsize=20)
    ax1.set_xticks([0, no_species-1])
    ax1.set_xticklabels([1, no_species ])
    # ax1.set_yticks([0,max(estimations[0].reference_sample_abundance.values())-1],[0,max(estimations[0].reference_sample_abundance.values())-1])

    plt.yticks([0, max(estimations[0].reference_sample_abundance.values()) - 1],
               [0, max(estimations[0].reference_sample_abundance.values()) - 1])

    no_species = len(estimations[1].reference_sample_abundance.values())
    ax2.plot(sorted(list(estimations[1].reference_sample_abundance.values()), reverse=True))
    ax2.fill_between(np.linspace(0, no_species, no_species, endpoint=False),
                     sorted(list(estimations[1].reference_sample_abundance.values()), reverse=True),
                     [0 for _ in range(no_species)], alpha=0.5)
    ax2.set_title(names[1] + " Noise", fontsize=24)
    ax2.set_xlabel("Species Rank", fontsize=20)
    #ax2.set_ylabel("Occurences", fontsize=20)
    ax2.set_xticks([0, no_species-1])
    ax2.set_xticklabels([1, no_species ])
    # ax2.set_yticks([0,max(estimations[0].reference_sample_abundance.values())-1],[0,max(estimations[0].reference_sample_abundance.values())-1])

    no_species = len(estimations[2].reference_sample_abundance.values())
    ax3.plot(sorted(list(estimations[2].reference_sample_abundance.values()), reverse=True))
    ax3.fill_between(np.linspace(0, no_species, no_species, endpoint=False),
                     sorted(list(estimations[2].reference_sample_abundance.values()), reverse=True),
                     [0 for _ in range(no_species)], alpha=0.5)
    ax3.set_title(names[2] + " Noise", fontsize=24)
    ax3.set_xlabel("Species Rank", fontsize=20)
    #ax3.set_ylabel("Occurences", fontsize=20)
    ax3.set_xticks([0, no_species-1])
    ax3.set_xticklabels([1, no_species ])
    # ax3.set_yticks([0,max(estimations[0].reference_sample_abundance.values())-1],[0,max(estimations[0].reference_sample_abundance.values())-1])

    no_species = len(estimations[3].reference_sample_abundance.values())
    ax4.plot(sorted(list(estimations[3].reference_sample_abundance.values()), reverse=True))
    ax4.fill_between(np.linspace(0, no_species, no_species, endpoint=False),
                     sorted(list(estimations[3].reference_sample_abundance.values()), reverse=True),
                     [0 for _ in range(no_species)], alpha=0.5)
    ax4.set_title(names[3] + " Noise", fontsize=24)
    ax4.set_xlabel("Species Rank", fontsize=20)
    #ax4.set_ylabel("Occurences", fontsize=20)
    ax4.set_xticks([0, no_species-1])
    ax4.set_xticklabels([1, no_species])
    # ax4.set_yticks([0,max(estimations[0].reference_sample_abundance.values())-1],[0,max(estimations[0].reference_sample_abundance.values())-1])

    plt.tight_layout()
    plt.savefig("figures/noise_eval_" + est_id + ".pdf", format="pdf")
    plt.close()
    #plt.show()


def plot_all_stats(estimations, name, loc=[0, 20, 40, 60, 80, 100], labels=[0, 200, 400, 600, 800, 1000]):
    #plt.style.use('seaborn-v0_8-ticks')
    #plt.style.use('ggplot')

    #labels = estimations[0].observation_ids
    #print(labels)
    no_data_points = len(estimations[0].observation_ids)
    #no_repetitions = len(estimations)
    #no_observations_abundance = estimations[0].number_observations_abundance
    #no_observations_incidence = estimations[0].number_observations_incidence

    obs_stats = summarize_metric("observed_species_count", estimations)
    chao2_abundance_stats = summarize_metric("chao2_abundance", estimations)
    chao2_incidence_stats = summarize_metric("chao2_incidence", estimations)

    D1_abundance_sample_stats = summarize_metric("D1_sample_abundance", estimations)
    D1_incidence_sample_stats = summarize_metric("D1_sample_incidence", estimations)
    D1_abundance_estimated_stats = summarize_metric("D1_estimated_abundance", estimations)
    D1_incidence_estimated_stats = summarize_metric("D1_estimated_incidence", estimations)

    D2_abundance_sample_stats = summarize_metric("D2_sample_abundance", estimations)
    D2_incidence_sample_stats = summarize_metric("D2_sample_incidence", estimations)
    D2_abundance_estimated_stats = summarize_metric("D2_estimated_abundance", estimations)
    D2_incidence_estimated_stats = summarize_metric("D2_estimated_incidence", estimations)

    #singletons_abundance_stats = summarize_metric("Q1_abundance", estimations)
    #singletons_incidence_stats = summarize_metric("Q1_incidence", estimations)

    #doubletons_abundance_stats = summarize_metric("Q2_abundance", estimations)
    #doubletons_incidence_stats = summarize_metric("Q2_incidence", estimations)

    coverage_abundance_stats = summarize_metric("coverage_abundance", estimations)
    coverage_incidence_stats = summarize_metric("coverage_incidence", estimations)

    completeness_abundance_stats = summarize_metric("completeness_abundance", estimations)
    completeness_incidence_stats = summarize_metric("completeness_incidence", estimations)

    l90_abundance_stats = summarize_metric("l90_abundance", estimations)
    l90_incidence_stats = summarize_metric("l90_incidence", estimations)

    l95_abundance_stats = summarize_metric("l95_abundance", estimations)
    l95_incidence_stats = summarize_metric("l95_incidence", estimations)

    l99_abundance_stats = summarize_metric("l99_abundance", estimations)
    l99_incidence_stats = summarize_metric("l99_incidence", estimations)

    plt.rcParams['figure.figsize'] = [9, 5]
    plt.rcParams['xtick.labelsize'] = 20
    plt.rcParams['ytick.labelsize'] = 20

    # abundance plots
    # NOTE: we only collect stats after trace level, thus we plot abundance-based stats using incidence observation counts
    # as number of observations instead of abundance observation counts

    # all hill numbers as observed, singletons vs doubletons, completeness vs coverage
    # for each hill number, actual value vs estimated value
    # f, (ax1, ax2, ax3) = plt.subplots(nrows=1, ncols=3, sharey='col')
    # Hill number q=0 & q=1 & q=2
    # print(no_data_points, len(obs_stats["lower_ci"]), len(obs_stats["upper_ci"]))
    # print(list(obs_stats["lower_ci"]))
    # print(list(obs_stats["upper_ci"]))
    # #np.linspace(0, no_data_points, num=no_data_points).astype(float),
    # plt.fill_between(np.linspace(0, no_data_points, num=no_data_points), obs_stats["lower_ci"], obs_stats["upper_ci"],
    #                  alpha=0.4)
    # plt.plot(obs_stats["median"], label="q=0")
    # plt.fill_between(np.linspace(0, no_data_points, num=no_data_points),D1_abundance_sample_stats["lower_ci"],
    #                  D1_abundance_sample_stats["upper_ci"], alpha=0.4)
    # plt.plot(D1_abundance_sample_stats["median"], label="q=1")
    # plt.fill_between(np.linspace(0, no_data_points, num=no_data_points),D2_abundance_sample_stats["lower_ci"],
    #                  D2_abundance_sample_stats["upper_ci"], alpha=0.4)
    # plt.plot(D2_abundance_sample_stats["median"], label="q=2")
    # #plt.title("Sample-based Hill-number (Abundance)", fontsize=26)
    # plt.xlabel("Sample Size", fontsize=22)
    # plt.ylabel("Hill number", fontsize=22)
    # plt.xticks([0, 20, 40, 60, 80, 100], [0, 200, 400, 600, 800, 1000])
    # plt.legend()
    # plt.tight_layout()
    # plt.savefig("figures/" + name + "_sample_hill_abundance.pdf", format="pdf")
    # #plt.close()
    # plt.show()

    # # Singletons vs Doubletons
    # plt.fill_between(np.linspace(0, no_data_points, num=no_data_points).astype(float), singletons_abundance_stats["lower_ci"],
    #                  singletons_abundance_stats["upper_ci"], alpha=0.4)
    # plt.plot(singletons_abundance_stats["median"], label="Singletons")
    # plt.fill_between(np.linspace(0, no_data_points, num=no_data_points).astype(float), doubletons_abundance_stats["lower_ci"],
    #                  doubletons_abundance_stats["upper_ci"], alpha=0.4)
    # plt.plot(doubletons_abundance_stats["median"], label="Doubletons")
    # #plt.title("Singletons vs Doubletons (Abundance)", fontsize=26)
    # plt.xlabel("Sample Size", fontsize=22)
    # plt.ylabel("Counts", fontsize=22)
    # plt.xticks([0, 20, 40, 60, 80, 100], [0, 200, 400, 600, 800, 1000])
    # plt.legend()
    # plt.tight_layout()
    # plt.savefig("figures/" + name + "_q1_vs_q2_abundance.pdf", format="pdf")
    # plt.close()
    # #plt.show()

    # completeness vs Coverage
    plt.fill_between(np.linspace(0, no_data_points, num=no_data_points).astype(float), completeness_abundance_stats["lower_ci"],
                     completeness_abundance_stats["upper_ci"], alpha=0.4)
    plt.plot(completeness_abundance_stats["median"], label="Completeness")
    plt.fill_between(np.linspace(0, no_data_points, num=no_data_points).astype(float), coverage_abundance_stats["lower_ci"],
                     coverage_abundance_stats["upper_ci"], alpha=0.4)
    plt.plot(coverage_abundance_stats["median"], label="Coverage")
    #plt.title("Completeness & Coverage (Abundance)", fontsize=26)
    plt.xlabel("Sample Size", fontsize=24)
    plt.ylabel("Completeness", fontsize=24)
    plt.xticks(loc, labels)
    plt.legend(fontsize=20)
    plt.tight_layout()
    plt.savefig("figures/" + name + "_completeness_vs_coverage_abundance.pdf", format="pdf")
    plt.close()
    #plt.show()

    # incidence-based stats
    # plt.fill_between(np.linspace(0, no_data_points, num=no_data_points).astype(float), obs_stats["lower_ci"], obs_stats["upper_ci"],
    #                  alpha=0.4)
    # plt.plot(obs_stats["median"], label="q=0")
    # plt.fill_between(np.linspace(0, no_data_points, num=no_data_points).astype(float), D1_incidence_sample_stats["lower_ci"],
    #                  D1_incidence_sample_stats["upper_ci"], alpha=0.4)
    # plt.plot(D1_incidence_sample_stats["median"], label="q=1")
    # plt.fill_between(np.linspace(0, no_data_points, num=no_data_points).astype(float), D2_incidence_sample_stats["lower_ci"],
    #                  D2_incidence_sample_stats["upper_ci"], alpha=0.4)
    # plt.plot(D2_incidence_sample_stats["median"], label="q=2")
    # #plt.title("Sample-based Hill-number (Incidence)", fontsize=26)
    # plt.xlabel("Sample Size", fontsize=24)
    # plt.ylabel("Hill number", fontsize=24)
    # plt.xticks([0, 20, 40, 60, 80, 100], [0, 200, 400, 600, 800, 1000])
    # plt.legend()
    # plt.tight_layout()
    # plt.savefig("figures/" + name + "_sample_hill_incidence.pdf", format="pdf")
    # plt.close()
    # #plt.show()
    #
    # # Singletons vs Doubletons
    # plt.fill_between(np.linspace(0, no_data_points, num=no_data_points).astype(float), singletons_incidence_stats["lower_ci"],
    #                  singletons_incidence_stats["upper_ci"], alpha=0.4)
    # plt.plot(singletons_incidence_stats["median"], label="Singletons")
    # plt.fill_between(np.linspace(0, no_data_points, num=no_data_points).astype(float), doubletons_incidence_stats["lower_ci"],
    #                  doubletons_incidence_stats["upper_ci"], alpha=0.4)
    # plt.plot(doubletons_incidence_stats["median"], label="Doubletons")
    # #plt.title("Singletons vs Doubletons (Incidence)", fontsize=26)
    # plt.xlabel("Sample Size", fontsize=24)
    # plt.ylabel("Counts", fontsize=24)
    # plt.xticks([0, 20, 40, 60, 80, 100], [0, 200, 400, 600, 800, 1000])
    # plt.legend()
    # plt.tight_layout()
    # plt.savefig("figures/" + name + "_q1_vs_q2_incidence.pdf", format="pdf")
    # plt.close()
    # #plt.show()

    # completeness vs Coverage
    plt.fill_between(np.linspace(0, no_data_points, num=no_data_points).astype(float), completeness_incidence_stats["lower_ci"],
                     completeness_incidence_stats["upper_ci"], alpha=0.4)
    plt.plot(completeness_incidence_stats["median"], label="Completeness")
    plt.fill_between(np.linspace(0, no_data_points, num=no_data_points).astype(float), coverage_incidence_stats["lower_ci"],
                     coverage_incidence_stats["upper_ci"], alpha=0.4)
    plt.plot(coverage_incidence_stats["median"], label="Coverage")
    #plt.title("Completeness & Coverage (Incidence)", fontsize=26)
    plt.xlabel("Sample Size", fontsize=24)
    plt.ylabel("Completeness", fontsize=24)
    plt.xticks(loc, labels)
    plt.legend(fontsize=20)
    plt.tight_layout()
    plt.savefig("figures/" + name + "_completeness_vs_coverage_incidence.pdf", format="pdf")
    plt.close()
    #plt.show()



    plt.rcParams['figure.figsize'] = [3*WIDTH, 5]
    plt.rcParams['xtick.labelsize'] = 20
    plt.rcParams['ytick.labelsize'] = 20

    f, (ax1, ax2, ax3) = plt.subplots(nrows=1, ncols=3, sharey='all')


    ax1.fill_between(np.linspace(0, no_data_points, num=no_data_points).astype(float), obs_stats["lower_ci"], obs_stats["upper_ci"],
                     alpha=0.4)
    ax1.plot(obs_stats["median"], label="Observed")
    ax1.fill_between(np.linspace(0, no_data_points, num=no_data_points).astype(float), chao2_abundance_stats["lower_ci"],
                     chao2_abundance_stats["upper_ci"], alpha=0.4)
    ax1.plot(chao2_abundance_stats["median"], label="Estimated")
    ax1.set_title("q=0", fontsize=28)
    ax1.set_xlabel("Sample Size", fontsize=24)
    ax1.set_ylabel("Hill number", fontsize=24)
    ax1.set_xticks(loc)
    ax1.set_xticklabels(labels)
    ax1.legend(fontsize=20)

    ax2.fill_between(np.linspace(0, no_data_points, num=no_data_points).astype(float), D1_abundance_estimated_stats["lower_ci"],
                     D1_abundance_estimated_stats["upper_ci"], alpha=0.4)
    ax2.plot(D1_abundance_estimated_stats["median"], label="Observed")
    ax2.fill_between(np.linspace(0, no_data_points, num=no_data_points).astype(float), D1_abundance_sample_stats["lower_ci"],
                     D1_abundance_sample_stats["upper_ci"], alpha=0.4)
    ax2.plot(D1_abundance_sample_stats["median"], label="Estimated")
    ax2.set_title("q=1", fontsize=28)
    ax2.set_xlabel("Sample Size", fontsize=24)
    ax2.set_ylabel("Hill number", fontsize=24)
    ax2.set_xticks(loc)
    ax2.set_xticklabels(labels)
    ax2.legend(fontsize=20)

    ax3.fill_between(np.linspace(0, no_data_points, num=no_data_points).astype(float), D2_abundance_estimated_stats["lower_ci"],
                     D2_abundance_estimated_stats["upper_ci"], alpha=0.4)
    ax3.plot(D2_abundance_estimated_stats["median"], label="Observed")
    ax3.fill_between(np.linspace(0, no_data_points, num=no_data_points).astype(float), D2_abundance_sample_stats["lower_ci"],
                     D2_abundance_sample_stats["upper_ci"], alpha=0.4)
    ax3.plot(D2_abundance_sample_stats["median"], label="Estimated")
    ax3.set_title("q=2", fontsize=28)
    ax3.set_xlabel("Sample Size", fontsize=24)
    ax3.set_ylabel("Hill number", fontsize=24)
    ax3.set_xticks(loc)
    ax3.set_xticklabels(labels)
    ax3.legend(fontsize=20)

    plt.ylim(bottom=0)

    plt.tight_layout()
    plt.savefig("figures/" + name + "_estimations_abundance.pdf", format="pdf")
    plt.close()
    #plt.show()

    f, (ax4, ax5, ax6) = plt.subplots(nrows=1, ncols=3, sharey='all')

    ax4.fill_between(np.linspace(0, no_data_points, num=no_data_points).astype(float), obs_stats["lower_ci"], obs_stats["upper_ci"],
                     alpha=0.4)
    ax4.plot(obs_stats["median"], label="Observed")
    ax4.fill_between(np.linspace(0, no_data_points, num=no_data_points).astype(float), chao2_incidence_stats["lower_ci"],
                     chao2_incidence_stats["upper_ci"], alpha=0.4)
    ax4.plot(chao2_incidence_stats["median"], label="Estimated")
    ax4.set_title("q=0", fontsize=28)
    ax4.set_xlabel("Sample Size", fontsize=24)
    ax4.set_ylabel("Hill number", fontsize=24)
    ax4.set_xticks(loc)
    ax4.set_xticklabels(labels)
    ax4.legend(fontsize=20)

    ax5.fill_between(np.linspace(0, no_data_points, num=no_data_points).astype(float), D1_incidence_sample_stats["lower_ci"],
                     D1_incidence_sample_stats["upper_ci"], alpha=0.4)
    ax5.plot(D1_incidence_sample_stats["median"], label="Observed")
    ax5.fill_between(np.linspace(0, no_data_points, num=no_data_points).astype(float), D1_incidence_estimated_stats["lower_ci"],
                     D1_incidence_estimated_stats["upper_ci"], alpha=0.4)
    ax5.plot(D1_incidence_estimated_stats["median"], label="Estimated")
    ax5.set_title("q=1", fontsize=28)
    ax5.set_xlabel("Sample Size", fontsize=24)
    ax5.set_ylabel("Hill number", fontsize=24)
    ax5.set_xticks(loc)
    ax5.set_xticklabels(labels)
    ax5.legend(fontsize=20)

    ax6.fill_between(np.linspace(0, no_data_points, num=no_data_points).astype(float), D2_incidence_sample_stats["lower_ci"],
                     D2_incidence_sample_stats["upper_ci"], alpha=0.4)
    ax6.plot(D2_incidence_sample_stats["median"], label="Observed")
    ax6.fill_between(np.linspace(0, no_data_points, num=no_data_points).astype(float), D2_incidence_estimated_stats["lower_ci"],
                     D2_incidence_estimated_stats["upper_ci"], alpha=0.4)
    ax6.plot(D2_incidence_estimated_stats["median"], label="Estimated")
    ax6.set_title("q=2", fontsize=28)
    ax6.set_xlabel("Sample Size", fontsize=24)
    ax6.set_ylabel("Hill number", fontsize=24)
    ax6.set_xticks(loc)
    ax6.set_xticklabels(labels)
    ax6.legend(fontsize=20)

    plt.ylim(bottom=0)

    plt.tight_layout()
    plt.savefig("figures/" + name + "_estimations_incidence.pdf", format="pdf")
    plt.close()
    #plt.show()

    plt.rcParams['figure.figsize'] = [WIDTH, 5]

    # f, (ax1,ax2) = plt.subplots(nrows=1, ncols=2)
    plt.fill_between(np.linspace(0, no_data_points, num=no_data_points).astype(float), l99_abundance_stats["lower_ci"],
                     l99_abundance_stats["upper_ci"], alpha=0.4)
    plt.plot(l99_abundance_stats["median"], label="l=.99")
    plt.fill_between(np.linspace(0, no_data_points, num=no_data_points).astype(float), l95_abundance_stats["lower_ci"],
                     l95_abundance_stats["upper_ci"], alpha=0.4)
    plt.plot(l95_abundance_stats["median"], label="l=.95")
    plt.fill_between(np.linspace(0, no_data_points, num=no_data_points).astype(float), l90_abundance_stats["lower_ci"],
                     l90_abundance_stats["upper_ci"], alpha=0.4)
    plt.plot(l90_abundance_stats["median"], label="l=.90")
    #plt.title("Expected Sampling Effort (Incidence)", fontsize=26)
    plt.xlabel("Sample Size", fontsize=24)
    plt.ylabel("Expected Sampling Effort", fontsize=24)
    plt.xticks(loc, labels)
    plt.legend(fontsize=20)

    plt.tight_layout()
    plt.savefig("figures/" + name + "_sampling_effort_abundance.pdf", format="pdf")
    plt.close()
    #plt.show()

    plt.fill_between(np.linspace(0, no_data_points, num=no_data_points).astype(float), l99_incidence_stats["lower_ci"],
                     l99_incidence_stats["upper_ci"], alpha=0.4)
    plt.plot(l99_incidence_stats["median"], label="l=.99")
    plt.fill_between(np.linspace(0, no_data_points, num=no_data_points).astype(float), l95_incidence_stats["lower_ci"],
                     l95_incidence_stats["upper_ci"], alpha=0.4)
    plt.plot(l95_incidence_stats["median"], label="l=.95")
    plt.fill_between(np.linspace(0, no_data_points, num=no_data_points).astype(float), l90_incidence_stats["lower_ci"],
                     l90_incidence_stats["upper_ci"], alpha=0.4)
    plt.plot(l90_incidence_stats["median"], label="l=.90")
    #plt.title("Expected Sampling Effort (Incidence)", fontsize=26)
    plt.xlabel("Sample Size", fontsize=24)
    plt.ylabel("Expected Sampling Effort", fontsize=24)
    plt.xticks(loc, labels)
    # plt.set_xticklabels()
    plt.legend(fontsize=20)

    plt.tight_layout()
    plt.savefig("figures/" + name + "_sampling_effort_incidence.pdf", format="pdf")
    plt.close()
    #plt.show()


import math
from random import sample

import mpmath
from numpy import euler_gamma
from scipy.special import digamma

from cachetools import cached


#TODO unify incidence and abundance-based methods in one function
def get_incidence_count(obs_species_counts: dict, i: int) -> int:
    """
    returns the number of species, that have an incidence count of i
    :param obs_species_counts: the species with corresponding incidence counts
    :param i: the incidence count
    :return: the number of species with incidence count i
    """
    return list(obs_species_counts.values()).count(i)


def get_singletons(obs_species_counts: dict) -> int:
    """
    returns the number of singletons species, i.e. those species that have an incidence count of 1
    :param obs_species_counts: the species with corresponding incidence counts
    :return: the number of species with incidence count 1
    """
    return list(obs_species_counts.values()).count(1)


def get_doubletons(obs_species_counts: dict) -> int:
    """
    returns the number of doubleton species, i.e. those species that have an incidence count of 2
    :param obs_species_counts: the species with corresponding incidence counts
    :return: the number of species with incidence count 2
    """
    return list(obs_species_counts.values()).count(2)


def get_number_observed_species(obs_species_counts: dict) -> int:
    """
    returns the number of observed species
    :param obs_species_counts: the species with corresponding incidence counts
    :return: the number of observed species
    """
    return len(obs_species_counts.keys())


def get_total_species_count(obs_species_counts):
    """
    returns the total number of species incidences, i.e. the sum of all species incidences in the reference sample
    :param obs_species_counts: the species with corresponding incidence counts
    :return: the sum of species incidences
    """
    return sum(obs_species_counts.values())


def hill_number(d: int, obs_species_counts: dict) -> float:
    """
    computes sample-based Hill number of order d for the reference sample.
    D=0 - species richness
    D=1 - Exponential of Shannon entropy
    D=2 - Simpson Diversity Index
    :param d: the order of the Hill number
    :param obs_species_counts: the species with corresponding incidence counts
    :return: the sample-based Hill number of order d
    """
    # sample-based species richness
    if d == 0:
        return get_number_observed_species(obs_species_counts)
    # sample-based exponential Shannon diversity
    if d == 1:
        return entropy_exp(obs_species_counts)
    # sample-based Simpson diversity
    if d == 2:
        return simpson_diversity(obs_species_counts)


def entropy_exp(obs_species_counts: dict) -> float:
    """
    computes the exponential of Shannon entropy
    :param obs_species_counts: the species with corresponding incidence counts
    :return: the exponential of Shannon entropy
    """
    total_species_count = get_total_species_count(obs_species_counts)
    # total = sum([obs_species_counts.values])
    return math.exp(-1 * sum(
        [x / total_species_count * math.log(x / total_species_count) for x in obs_species_counts.values()]))


def simpson_diversity(obs_species_counts: dict) -> float:
    """
    computes the Simpson diversity index
    :param obs_species_counts: the species with corresponding incidence counts
    :return: the Simpson diversity index
    """
    total_species_count = get_total_species_count(obs_species_counts)

    a = sum([(x / total_species_count) ** 2 for x in obs_species_counts.values()])
    # TODO check if return 1 is reasonable
    return a ** (1 / (1 - 2)) if a > 0 else 1


'''
Calculate asymptotic Hill number of order d for a reference sample
d=0 Species Richness
d=1 Exponential of Shannon Entropy
d=2 Simpson Diversity Index
'''


def hill_number_asymptotic(d: int, obs_species_counts: dict, sample_size: int, abundance: bool = True) -> float:
    """
    computes asymptotic Hill number of order d for the reference sample, for either abundance data or incidence data.
    D=0 - asymptotic species richness
    D=1 - asymptotic exponential of Shannon entropy
    D=2 - asymptotic Simpson diversity index
    :param d: the order of the Hill number
    :param obs_species_counts: the species with corresponding incidence counts
    :param sample_size: the sample size associated with the species incidence counts
    :param abundance: flag indicating the data type. Setting this 'True' indicates abundance-based data,
    setting this 'False' indicates incidence-based data
    :return: the asymptotic Hill number of order d
    """
    # asymptotic species richness
    # for species richness, there is no differentiation between abundance and incidence
    if d == 0:
        return estimate_species_richness_chao(obs_species_counts)
    # asymptotic Shannon entropy
    if d == 1:
        if abundance:
            return estimate_exp_shannon_entropy_abundance(obs_species_counts, sample_size)
        # incidence
        else:
            return estimate_exp_shannon_entropy_incidence(obs_species_counts, sample_size)
    # asymptotic Simpson diversity
    if d == 2:
        if abundance:
            return estimate_simpson_diversity_abundance(obs_species_counts, sample_size)
        # incidence
        else:
            return estimate_simpson_diversity_incidence(obs_species_counts, sample_size)


def estimate_species_richness_chao(obs_species_counts: dict) -> float:
    """
    computes the asymptotic(=estimated) species richness using the Chao1 estimator(for abundance data)
    or Chao2 estimator (for incidence data)
    :param obs_species_counts: the species with corresponding incidence counts
    :return: the estimated species richness
    """
    obs_species_count = get_number_observed_species(obs_species_counts)
    f_1 = get_singletons(obs_species_counts)
    f_2 = get_doubletons(obs_species_counts)

    if f_2 != 0:
        return obs_species_count + f_1 ** 2 / (2 * f_2)
    else:
        return obs_species_count + f_1 * (f_1 - 1) / 2


def estimate_species_richness_chao_corrected(obs_species_counts: dict) -> float:
    """
    computes the asymptotic(=estimated) species richness using the Chao1 estimator(for abundance data)
    or Chao2 estimator (for incidence data). Includes a correction term for very small sample sizes
    :param obs_species_counts: the species with corresponding incidence counts
    :return: the estimated species richness
    """
    obs_species_count = get_number_observed_species(obs_species_counts)
    f_1 = get_singletons(obs_species_counts)
    f_2 = get_doubletons(obs_species_counts)

    if f_2 != 0:
        return ((obs_species_count - 1) / obs_species_count) * obs_species_count + f_1 ** 2 / (2 * f_2)
    else:
        return ((obs_species_count - 1) / obs_species_count) * obs_species_count + f_1 * (f_1 - 1) / 2


def estimate_exp_shannon_entropy_abundance(obs_species_counts: dict, sample_size: int) -> float:
    """
    computes the asymptotic(=estimated) exponential of Shannon entropy for abundance-based data
    :param obs_species_counts: the species with corresponding incidence counts
    :param sample_size: the sample size associated with the species incidence counts
    :return: the estimated exponential of Shannon entropy
    """
    if sum(obs_species_counts.values())==0 or sample_size==0:
        return 0.0
    return math.exp(estimate_entropy(obs_species_counts, sample_size))


def estimate_exp_shannon_entropy_incidence(obs_species_counts: dict, sample_size: int) -> float:
    """
    computes the asymptotic(=estimated) exponential of Shannon entropy for incidence-based data
    :param obs_species_counts: the species with corresponding incidence counts
    :param sample_size: the sample size associated with the species incidence counts
    :return: the estimated exponential of Shannon entropy
    """
    # term h_o is structurally equivalent to abundance based entropy estimation, see eq H7 in appendix H of Hill number paper
    u = sum(obs_species_counts.values())
    h_o = estimate_entropy(obs_species_counts, sample_size)
    if u == 0:
        return 0.0
    #print(h_o, u, sample_size)
    #print(">>>"+str(math.exp((sample_size / u) * h_o + math.log(u / sample_size))))
    return math.exp((sample_size / u) * h_o + math.log(u / sample_size))


def estimate_entropy(obs_species_counts: dict, sample_size: int) -> float:
    """
    computes the estimated Shannon entropy
    :param obs_species_counts: the species with corresponding incidence counts
    :param sample_size: the sample size associated with the species incidence counts
    :return: the estimated exponential of Shannon entropy
    """
    if sample_size <= 1:
        return 0

    f_1 = get_singletons(obs_species_counts)
    f_2 = get_doubletons(obs_species_counts)

    entropy_known_species = 0

    for x_i in obs_species_counts.values():
        if x_i <= sample_size - 1:
            norm_factor = x_i / sample_size

            #decompose sum(1/x_i,...,1/sample_size) to um(1/1,...,1/sample_size)-sum(1/1,...,1/x_i-1)
            entropy_known_species = entropy_known_species + norm_factor * (harmonic(sample_size) - harmonic(x_i-1))
            #entropy_known_species = entropy_known_species + norm_factor * (mpmath.harmonic(sample_size) - mpmath.harmonic(x_i-1))
            #entropy_known_species = entropy_known_species + norm_factor * sum([1 / k for k in range(x_i, sample_size)])

    #print(f_2,f_1, sample_size, sum(obs_species_counts.values()))
    a = 0
    if f_2 > 0:
        a = (2 * f_2) / ((sample_size - 1) * f_1 + 2 * f_2)
        #a = (2 * f_2) / ( f_1 + 2 * f_2)
    elif f_2 == 0 and f_1 > 0:
        a = 2 / ((sample_size - 1) * (f_1 - 1) + 2)
    else:
        a = 1
    #print(f_1, f_2, a, sample_size)
    entropy_unknown_species = 0
    if f_1==1 and f_2 >= 20:
        return entropy_known_species
    # TODO rethink if this is really necessary
    #if a == 1:
    #    return entropy_known_species
    #(((1 - a) ** (-sample_size + 1)))
    #print(0 ** 690)
    #print(((1 - a) ** (sample_size - 1)), sample_size, a)
    #print(entropy_known_species, a, f_1, f_2, sample_size)
    #print(a, f_1, f_2, sample_size)
    if a==1:
        return entropy_known_species
        #entropy_unknown_species = (f_1 / sample_size) * sum([(1 / r) for r in range(1, sample_size)])
        #return entropy_known_species + entropy_unknown_species
    ###((1-a) ** (-sample_size+1))
    #else:
    else:
        entropy_unknown_species = (f_1 / sample_size) * (1-a) ** (-sample_size+1) * (
                -math.log(a) - sum([1/r * ((1 - a) ** r) for r in range(1, sample_size)]))
        #print(entropy_known_species, entropy_unknown_species)
        #print(entropy_known_species, entropy_unknown_species, (f_1 / sample_size), ((1-a) ** (-sample_size+1)), -math.log(a), sum([(1 / r) * ((1 - a) ** r) for r in range(1, sample_size)]))
        #print("TEST "+str(entropy_known_species + entropy_unknown_species))
        #print(-math.log(a), sum([(1 / r) * ((1 - a) ** r) for r in range(1, sample_size)]), -math.log(a) - sum([(1 / r) * ((1 - a) ** r) for r in range(1, sample_size)]))
        return entropy_known_species + entropy_unknown_species


def harmonic(n):
    """Returns an (approximate) value of n-th harmonic number.
    If n>100, use an efficient approximation using the digamma function instead
    http://en.wikipedia.org/wiki/Harmonic_number
    taken from: https://stackoverflow.com/questions/404346/python-program-to-calculate-harmonic-series
     """
    if n <= 100:
        return sum(1/k for k in range(1,n+1))
    else:
        return digamma(n + 1) + euler_gamma


def estimate_simpson_diversity_abundance(obs_species_counts: dict, sample_size: int) -> float:
    """
    computes the asymptotic(=estimated) Simpson diversity for abundance-based data
    :param obs_species_counts: the species with corresponding incidence counts
    :param sample_size: the sample size associated with the species incidence counts
    :return: the estimated Simpson diversity
    """
    # TODO make this understandable
    denom = 0
    for x_i in obs_species_counts.values():
        if x_i >= 2:
            denom = denom + (x_i * (x_i - 1))
    if denom == 0:
        return 0
    return (sample_size * (sample_size - 1)) / denom


def estimate_simpson_diversity_incidence(obs_species_counts: dict, sample_size: int) -> float:
    """
    computes the asymptotic(=estimated) Simpson diversity for incidence-based data
    :param obs_species_counts: the species with corresponding incidence counts
    :param sample_size: the sample size associated with the species incidence counts
    :return: the estimated Simpson diversity
    """
    # TODO make this understandable
    u = get_total_species_count(obs_species_counts)
    s = 0
    if u == 0:
        return 0.0
    nom = ((1 - (1 / sample_size)) * u) ** 2

    for y_i in obs_species_counts.values():
        if y_i > 1:
            #    s = s + (sample_size ** 2 * y_i ** 2) / (u ** 2 * sample_size ** 2)
            s = s + (y_i * (y_i - 1))
    if s == 0:
        return 0
    # return s ** (1 / (1 - 2))
    return nom / s


def completeness(obs_species_counts: dict) -> float:
    """
    computes the completeness of the sample data. A value of '1' indicates full completeness,
    whereas as value of '0' indicates total incompleteness
    :param obs_species_counts: the species with corresponding incidence counts
    :return: the estimated completeness
    """
    obs_species_count = get_number_observed_species(obs_species_counts)
    s_P = estimate_species_richness_chao(obs_species_counts)
    if s_P == 0:
        return 0

    return obs_species_count / s_P


def coverage(obs_species_counts: dict, sample_size: int) -> float:
    """
    computes the coverage of the sample data. A value of '1' indicates full coverage,
    whereas as value of '0' indicates no coverage
    :param obs_species_counts: the species with corresponding incidence counts
    :param sample_size: the sample size associated with the species incidence counts
    :return: the estimated coverage
    """
    f_1 = get_singletons(obs_species_counts)
    f_2 = get_doubletons(obs_species_counts)
    Y = get_total_species_count(obs_species_counts)

    if sample_size == 0:
        return 0
    if f_2 == 0 and sample_size == 1:
        return 0
    if f_1 == 0 and f_2 == 0:
        return 1

    return 1 - f_1 / Y * (((sample_size - 1) * f_1) / ((sample_size - 1) * f_1 + 2 * f_2))


def sampling_effort_abundance(n: float, obs_species_counts: dict, sample_size: int) -> float:
    """
    computes the expected additional sampling effort needed to reach target completeness l for abundance data.
    If f exceeds the current completeness, this function returns 0
    :param n: desired target completeness
    :param obs_species_counts: the species with corresponding incidence counts
    :param sample_size: the sample size associated with the species incidence counts
    :return: the expected additional sampling effort
    """
    comp = completeness(obs_species_counts)
    f_1 = get_singletons(obs_species_counts)
    f_2 = get_doubletons(obs_species_counts)

    if n <= comp:
        return 0
    if f_2 == 0:
        return 0

    obs_species_count = estimate_species_richness_chao(obs_species_counts)
    #get_number_observed_species(obs_species_counts))

    s_P = 0
    if f_2 != 0:
        s_P = f_1 ** 2 / (2 * f_2)
    else:
        s_P = f_1 * (f_1 - 1) / 2

    return ((sample_size * f_1) / (2 * f_2)) * math.log(s_P / ((1 - n) * (s_P + obs_species_count)))


def sampling_effort_incidence(n: float, obs_species_counts: dict, sample_size: int) -> float:
    """
    computes the expected additional sampling effort needed to reach target completeness l for incidence data.
    If f exceeds the current completeness, this function returns 0
    :param n: desired target completeness
    :param obs_species_counts: the species with corresponding incidence counts
    :param sample_size: the sample size associated with the species incidence counts
    :return: the expected additional sampling effort
    """
    comp = completeness(obs_species_counts)
    f_1 = get_singletons(obs_species_counts)
    f_2 = get_doubletons(obs_species_counts)
    if n <= comp:
        return 0
    #if f_2 == 0:
    #    return 0
    if sample_size==1:
        return 0
    if f_1 == 0:
        return 0

    # for small sample sizes, correction term is introduced, otherwise math error
    #obs_species_count = estimate_species_richness_chao(obs_species_counts)
    obs_species_count = get_number_observed_species(obs_species_counts)

    s_P = 0
    if f_2 != 0:
        s_P = obs_species_count + (1 - 1 / sample_size) * f_1 ** 2 / (2 * f_2)
    else:
        s_P = obs_species_count + (1 - 1 / sample_size) * f_1 * (f_1 - 1) / 2

    #s_P = obs_species_count + (1 - 1 / sample_size) * f_1 ** 2 / (2 * f_2)


    #TODO double check if this is indeed correct
    #should f_2 be 0, technically assessment is not possible, thus we treat it as if one doubletons remained.
    # Thus results are approximative in this case

    if f_2!=0:
        nom1 = (sample_size / (sample_size - 1))
        nom2 = ((2 * f_2) / (f_1 ** 2))
        nom3 = (n * s_P - obs_species_count)
        nom = (math.log(1 - nom1 * nom2 * nom3))
        denominator = (math.log(1 - ((2 * f_2) / ((sample_size - 1) * f_1 + 2 * f_2))))
        return nom / denominator
    else:
        nom1 = (sample_size / (sample_size - 1))
        nom2 = ((2 * 1) / (f_1 ** 2))
        nom3 = (n * s_P - obs_species_count)
        nom = (math.log(1 - nom1 * nom2 * nom3))
        denominator = (math.log(1 - ((2 * 1) / ((sample_size - 1) * f_1 + 2 * 1))))
        return nom / denominator
    #return final



from audioop import reverse
from cProfile import label
from enum import Enum

import matplotlib.pyplot as plt
import numpy as np

from special4pm.estimation.species_estimator import SpeciesEstimator


class plot_params(Enum):
    WIDTH = 9


#TODO add bootstrap interval to asymptotic analysis
def plot_rank_abundance(estimator: SpeciesEstimator, species_id: str, abundance: bool = False, save_to=None):
    #plt.style.use('seaborn-v0_8-ticks')

    plt.rcParams['figure.figsize'] = [9, 3.5]
    plt.rcParams['xtick.labelsize'] = 20
    plt.rcParams['ytick.labelsize'] = 20

    reference_sample = estimator.metrics[species_id].reference_sample_abundance if abundance \
        else estimator.metrics[species_id].reference_sample_incidence
    species_count = sum(reference_sample.values())
    reference_values_sorted = sorted(list(reference_sample.values()), reverse=True)# if absolute else [x/species_count for x in sorted(list(reference_sample.values()), reverse=True)]
    no_species = len(reference_sample)

    plt.fill_between(np.linspace(0, no_species, no_species, endpoint=False),
                     reference_values_sorted,
                     [0 for _ in range(no_species)], alpha=0.4)
    plt.plot(reference_values_sorted)
    plt.xlabel("Species Rank", fontsize=24)
    plt.ylabel("Occurrences", fontsize=24)
    #if not absolute:
    #    plt.ylim(0,1.1)
    plt.xticks([0, no_species - 1], [1, no_species])
    plt.yticks([0, max(reference_values_sorted)],
               [0, max(reference_values_sorted)])
    #if not absolute:
    #    plt.yticks([0, max(reference_values_sorted)],
    #               [0, '%.2f'%(max(reference_values_sorted))])


    plt.tight_layout()
    if save_to is None:
        plt.show()
    else:
        plt.savefig(save_to, format="pdf")
    plt.close()
        #plt.show()
    #plt.close()


#TODO properly unify function calls - the following three diversity functions can be unified into one function
def plot_diversity_sample_vs_estimate(estimator: SpeciesEstimator, species_id: str, metrics: [str],
                                      abundance: bool = False, save_to=None):
    '''
    Plots the obtained time series of sample-based diversity vs asymptotic diversity
    :param save_to:
    :param estimator:
    :param species_id:
    :param metrics:
    :param abundance:
    :return:
    '''
    plt.rcParams['figure.figsize'] = [3 * 9, 5]
    plt.rcParams['xtick.labelsize'] = 20
    plt.rcParams['ytick.labelsize'] = 20
    f, (ax1, ax2, ax3) = plt.subplots(nrows=1, ncols=3, sharey="row", layout="constrained")

    for metric, ax, title in zip(metrics, (ax1, ax2, ax3), ("D0", "D1", "D2")):
        key_sample = "abundance_sample_" + metric if abundance else "incidence_sample_" + metric
        key_estimate = "abundance_estimate_" + metric if abundance else "incidence_estimate_" + metric

        series_sample = estimator.metrics[species_id][key_sample] if abundance else estimator.metrics[species_id][
            key_sample]
        series_estimate = estimator.metrics[species_id][key_estimate] if abundance else estimator.metrics[species_id][
            key_estimate]

        series_observations_ids = estimator.metrics[species_id]["abundance_no_observations"] if abundance else \
            estimator.metrics[species_id][
                "incidence_no_observations"]

        no_data_points = len(series_sample)
        series_ticks = estimator.metrics[species_id]["abundance_no_observations"] if abundance else \
            estimator.metrics[species_id]["incidence_no_observations"]

        ax.set_xticks([0, no_data_points], [0, series_observations_ids[-1]])
        ax.plot(series_sample, label="observed")
        ax.plot(series_estimate, label="estimated")
        ax.set_title(title, fontsize=24)
        ax.set_xlabel("Sample Size", fontsize=24)
        ax.set_ylabel("Diversity", fontsize=24)
        ax.legend(fontsize=20)
    #plt.tight_layout()
    if save_to is None:
        plt.show()
    else:
        plt.savefig(save_to, format="pdf")
    plt.close()

def plot_diversity_sample_ccurve(estimator: SpeciesEstimator, species_id: str, metric: str,
                          abundance: bool = False, save_to=None):
    '''
    Plots the time series of the specified sample_based diversity metric, adding the asymptotic diversity as an indicator
    :param estimator:
    :param species_id:
    :param metric:
    :param abundance:
    :return:
    '''
    plt.rcParams['figure.figsize'] = [5, 4]
    plt.rcParams['xtick.labelsize'] = 20
    plt.rcParams['ytick.labelsize'] = 20
    f = plt.figure(layout="constrained")

    #for metric, ax, title in zip(metrics, (ax1, ax2, ax3), ("D0","D1","d2")):
    key_sample = "abundance_sample_" + metric if abundance else "incidence_sample_" + metric
    key_estimate = "abundance_estimate_" + metric if abundance else "incidence_estimate_" + metric

    series_sample = estimator.metrics[species_id][key_sample] if abundance else estimator.metrics[species_id][
        key_sample]
    series_estimate = estimator.metrics[species_id][key_estimate] if abundance else estimator.metrics[species_id][
        key_estimate]

    series_observations_ids = estimator.metrics[species_id]["abundance_no_observations"] if abundance else \
        estimator.metrics[species_id][
            "incidence_no_observations"]

    no_data_points = len(series_sample)
    series_ticks = estimator.metrics[species_id]["abundance_no_observations"] if abundance else \
        estimator.metrics[species_id]["incidence_no_observations"]

    plt.xticks([0, no_data_points], [0, series_observations_ids[-1]])

    plt.plot(series_sample, label=metric)
    plt.title(metric, fontsize=24)
    plt.xlabel("Sample Size", fontsize=24)
    plt.ylabel("Diversity", fontsize=24)
    plt.plot(no_data_points, series_estimate[-1], 'o', c="grey")
    #plt.annotate(metric + "=" + str(series_estimate[-1]), (no_data_points, series_estimate[-1]))
    #plt.tight_layout()
    if save_to is None:
        plt.show()
    else:
        plt.savefig(save_to, format="pdf")
        plt.show()


def plot_diversity_profile(estimator, species_id: str, metrics: [str] = ["d0","d1","d2"],
                              abundance: bool = False, save_to=None):
    if type(estimator) is not list: estimator = [estimator ]
    '''
    Plots all sample-based diversity metrics with their asymptotic diversity
    :param save_to:
    :param estimator:
    :param species_id:
    :param metrics:
    :param abundance:
    :return:
    '''
    plt.rcParams['figure.figsize'] = [5, 4]
    plt.rcParams['xtick.labelsize'] = 20
    plt.rcParams['ytick.labelsize'] = 20
    f = plt.figure(layout="constrained")
    for e in estimator:
        for metric, title in zip(metrics, ("D0", "D1", "D2")):
            key_sample = "abundance_sample_" + metric if abundance else "incidence_sample_" + metric
            key_estimate = "abundance_estimate_" + metric if abundance else "incidence_estimate_" + metric
    
            series_sample = e.metrics[species_id][key_sample] if abundance else e.metrics[species_id][
                key_sample]
            series_estimate = e.metrics[species_id][key_estimate] if abundance else e.metrics[species_id][
                key_estimate]
    
            series_observations_ids = e.metrics[species_id]["abundance_no_observations"] if abundance else \
                e.metrics[species_id][
                    "incidence_no_observations"]
    
            no_data_points = len(series_sample)
            series_ticks = e.metrics[species_id]["abundance_no_observations"] if abundance else \
                e.metrics[species_id]["incidence_no_observations"]
    
            plt.xticks([0, no_data_points], [0, series_observations_ids[-1]])
    
            plt.plot(series_sample, label=metric)
            plt.plot(no_data_points-1, series_estimate[-1], 'o', c="darkgrey")
        #plt.annotate(metric + "=" + str(series_estimate[-1]), (no_data_points, series_estimate[-1]))
    plt.xlabel("Sample Size", fontsize=24)
    plt.ylabel("Diversity", fontsize=24)

    plt.tight_layout()
    if save_to is None:
        plt.show()
    else:
        plt.savefig(save_to, format="pdf")
    plt.close()

def plot_diversity_profile_asymptotic(estimator, species_id: str,
                           abundance: bool = False, save_to = None):
    '''
    Plots the asymptotic diversity profile
    :param save_to:
    :param estimator:
    :param species_id:
    :param abundance:
    :return:
    '''
    if type(estimator) is not list: estimator = [estimator ]

    plt.rcParams['figure.figsize'] = [5, 4]
    plt.rcParams['xtick.labelsize'] = 20
    plt.rcParams['ytick.labelsize'] = 20
    f = plt.figure(layout="constrained")

    key = "abundance_estimate_" if abundance else "incidence_estimate_"
    i=1
    for e in estimator:
        profile = [e.metrics[species_id][key + "d0"][-1],
                   e.metrics[species_id][key + "d1"][-1],
                   e.metrics[species_id][key + "d2"][-1]]
        plt.plot(profile, label=i)
        i = i+1

    plt.legend()
    plt.xticks([0, 1, 2], ["q=0", "q=1", "q=2"])
    plt.xlabel("Order q", fontsize=22)
    plt.ylabel("Diversity", fontsize=22)

    plt.tight_layout()
    if save_to is None:
        plt.show()
    else:
        plt.savefig(save_to, format="pdf")
        plt.show()


#TODO ensure y axis always goes up to 1.0
def plot_completeness_profile(estimator, species_id: str,
                              abundance: bool  = False, save_to = None):
    '''
    Plots the time series for both completeness and coverage
    :param save_to:
    :param estimator:
    :param species_id:
    :param abundance:
    :return:
    '''
    if type(estimator) is not list: estimator = [estimator ]

    plt.rcParams['figure.figsize'] = [5, 4]
    plt.rcParams['xtick.labelsize'] = 20
    plt.rcParams['ytick.labelsize'] = 20
    f = plt.figure(layout="constrained")

    key_completeness = "abundance_c0" if abundance else "incidence_c0"
    key_coverage = "abundance_c1" if abundance else "incidence_c1"

    for e in estimator:
        series_completeness = e.metrics[species_id][key_completeness] if abundance else \
            e.metrics[species_id][
                key_completeness]
        series_coverage = e.metrics[species_id][key_coverage] if abundance else e.metrics[species_id][
            key_coverage]

        no_data_points = len(e.metrics[species_id][key_completeness] if abundance else \
                                 e.metrics[species_id][
                                     key_completeness])

        series_observations_ids = e.metrics[species_id]["abundance_no_observations"] if abundance else \
            e.metrics[species_id][
                "incidence_no_observations"]

        no_data_points = len(series_completeness)
        series_ticks = e.metrics[species_id]["abundance_no_observations"] if abundance else \
            e.metrics[species_id]["incidence_no_observations"]

        plt.xticks([0, no_data_points], [0, series_observations_ids[-1]])

        plt.plot(series_completeness, label="Completeness")
        plt.plot(series_coverage, label="Coverage")
        plt.legend()
    #plt.title("Completeness Profile", fontsize=24)
    plt.ylim((0,1.1))
    plt.xlabel("Sample Size", fontsize=22)
    plt.ylabel("Completeness", fontsize=22)

    if save_to is None:
        plt.show()
    else:
        plt.savefig(save_to, format="pdf")
    plt.close()

def plot_expected_sampling_effort(estimator: SpeciesEstimator, species_id: str,
                                  abundance: bool = False, save_to = None):
    '''
    Plots the time series for expected sampling efforts
    :param save_to:
    :param estimator:
    :param species_id:
    :param abundance:
    :return:
    '''
    plt.rcParams['figure.figsize'] = [5, 4]
    plt.rcParams['xtick.labelsize'] = 20
    plt.rcParams['ytick.labelsize'] = 20
    f = plt.figure(layout="constrained")

    key = "abundance_l_" if abundance else "incidence_l_"
    for n in estimator.l_n:
        series_effort = estimator.metrics[species_id][key + str(n)] if abundance else estimator.metrics[species_id][
            key + str(n)]

        series_observations_ids = estimator.metrics[species_id]["abundance_no_observations"] if abundance else \
            estimator.metrics[species_id][
                "incidence_no_observations"]

        no_data_points = len(series_effort)
        series_ticks = estimator.metrics[species_id]["abundance_no_observations"] if abundance else \
            estimator.metrics[species_id]["incidence_no_observations"]

        plt.xticks([0, no_data_points], [0, series_observations_ids[-1]])

        plt.plot(series_effort, label="l=" + str(n))
    plt.legend()
    #plt.title("Completeness Profile", fontsize=24)
    plt.xlabel("Sample Size", fontsize=24)
    if save_to is None:
        plt.show()
    else:
        plt.savefig(save_to, format="pdf")
    plt.close()